"""Unit tests for findscu.py"""

import logging
import os
import subprocess
import sys
import time

import pytest

from pydicom import dcmread
from pydicom.uid import (
    ExplicitVRLittleEndian,
    ImplicitVRLittleEndian,
    DeflatedExplicitVRLittleEndian,
    ExplicitVRBigEndian,
)

from pynetdicom import (
    AE,
    evt,
    debug_logger,
    DEFAULT_TRANSFER_SYNTAXES,
    QueryRetrievePresentationContexts,
    BasicWorklistManagementPresentationContexts,
    UnifiedProcedurePresentationContexts,
)
from pynetdicom.sop_class import (
    Verification,
    PatientRootQueryRetrieveInformationModelFind,
    StudyRootQueryRetrieveInformationModelFind,
    PatientStudyOnlyQueryRetrieveInformationModelFind,
    ModalityWorklistInformationFind,
    UnifiedProcedureStepPull,
)


# debug_logger()


APP_DIR = os.path.join(os.path.dirname(__file__), "../")
APP_FILE = os.path.join(APP_DIR, "findscu", "findscu.py")
DATA_DIR = os.path.join(APP_DIR, "../", "tests", "dicom_files")
DATASET_FILE = os.path.join(DATA_DIR, "CTImageStorage.dcm")


def start_findscu(args):
    """Start the findscu.py app and return the process."""
    pargs = [sys.executable, APP_FILE, "localhost", "11112"] + [*args]
    return subprocess.Popen(pargs)


def start_findscu_scli(args):
    """Start the findscu app using CLI and return the process."""
    pargs = [sys.executable, "-m", "pynetdicom", "findscu", "localhost", "11112"] + [
        *args
    ]
    return subprocess.Popen(pargs)


class FindSCUBase:
    """Tests for findscu.py"""

    def setup_method(self):
        """Run prior to each test"""
        self.ae = None
        self.func = None

    def teardown_method(self):
        """Clear any active threads"""
        if self.ae:
            self.ae.shutdown()

    def test_default(self):
        """Test default settings."""
        events = []

        def handle_find(event):
            events.append(event)
            yield 0x0000, None

        def handle_release(event):
            events.append(event)

        handlers = [(evt.EVT_C_FIND, handle_find), (evt.EVT_RELEASED, handle_release)]

        self.ae = ae = AE()
        ae.acse_timeout = 5
        ae.dimse_timeout = 5
        ae.network_timeout = 5
        ae.supported_contexts = (
            QueryRetrievePresentationContexts
            + BasicWorklistManagementPresentationContexts
            + UnifiedProcedurePresentationContexts
        )
        scp = ae.start_server(("localhost", 11112), block=False, evt_handlers=handlers)

        p = self.func(["-k", "PatientName="])
        p.wait()
        assert p.returncode == 0

        scp.shutdown()

        assert events[0].event == evt.EVT_C_FIND
        assert events[0].identifier.PatientName == ""
        assert events[1].event == evt.EVT_RELEASED
        requestor = events[1].assoc.requestor
        assert "FINDSCU" == requestor.ae_title
        assert 16382 == requestor.maximum_length
        assert "ANY-SCP" == requestor.primitive.called_ae_title
        assert [] == requestor.extended_negotiation
        assert (1, 1) == requestor.asynchronous_operations
        assert {} == requestor.sop_class_common_extended
        assert {} == requestor.sop_class_extended
        assert requestor.user_identity == None
        cxs = requestor.primitive.presentation_context_definition_list
        assert len(cxs) == 19
        cxs = {cx.abstract_syntax: cx for cx in cxs}
        assert PatientRootQueryRetrieveInformationModelFind in cxs
        cx = cxs[PatientRootQueryRetrieveInformationModelFind]
        assert cx.transfer_syntax == DEFAULT_TRANSFER_SYNTAXES

    def test_no_peer(self, capfd):
        """Test trying to connect to non-existent host."""
        p = self.func(["-k", "PatientName="])
        p.wait()
        assert p.returncode == 1

        out, err = capfd.readouterr()
        assert "Association request failed: unable to connect to remote" in err
        assert "TCP Initialisation Error" in err

    def test_bad_input(self, capfd):
        """Test being unable to read the input file."""
        p = self.func(["-f", "no-such-file.dcm"])
        p.wait()
        assert p.returncode == 1

        out, err = capfd.readouterr()
        assert "Cannot read input file no-such-file.dcm" in err

    def test_flag_version(self, capfd):
        """Test --version flag."""
        p = self.func(["--version"])
        p.wait()
        assert p.returncode == 0

        out, err = capfd.readouterr()
        assert "findscu.py v" in out

    def test_flag_quiet(self, capfd):
        """Test --quiet flag."""
        self.ae = ae = AE()
        ae.acse_timeout = 5
        ae.dimse_timeout = 5
        ae.network_timeout = 5
        ae.add_supported_context(Verification)
        scp = ae.start_server(("localhost", 11112), block=False)

        p = self.func(["-q", "-k", "PatientName="])
        p.wait()
        assert p.returncode == 1

        out, err = capfd.readouterr()
        assert out == err == ""

        scp.shutdown()

    def test_flag_verbose(self, capfd):
        """Test --verbose flag."""

        def handle_find(event):
            yield 0x0000, None

        handlers = [
            (evt.EVT_C_FIND, handle_find),
        ]

        self.ae = ae = AE()
        ae.acse_timeout = 5
        ae.dimse_timeout = 5
        ae.network_timeout = 5
        ae.supported_contexts = (
            QueryRetrievePresentationContexts
            + BasicWorklistManagementPresentationContexts
        )
        scp = ae.start_server(("localhost", 11112), block=False, evt_handlers=handlers)

        p = self.func(["-v", "-k", "PatientName="])
        p.wait()
        assert p.returncode == 0

        out, err = capfd.readouterr()
        assert "Requesting Association" in err
        assert "Association Accepted" in err
        assert "Sending Find Request" in err
        assert "Find SCP Result" in err
        assert "Releasing Association" in err
        assert "Accept Parameters" not in err

        scp.shutdown()

    def test_flag_debug(self, capfd):
        """Test --debug flag."""

        def handle_find(event):
            yield 0x0000, None

        handlers = [
            (evt.EVT_C_FIND, handle_find),
        ]

        self.ae = ae = AE()
        ae.acse_timeout = 5
        ae.dimse_timeout = 5
        ae.network_timeout = 5
        ae.supported_contexts = (
            QueryRetrievePresentationContexts
            + BasicWorklistManagementPresentationContexts
        )
        scp = ae.start_server(("localhost", 11112), block=False, evt_handlers=handlers)

        p = self.func(["-d", "-k", "PatientName="])
        p.wait()
        assert p.returncode == 0

        out, err = capfd.readouterr()
        assert "Releasing Association" in err
        assert "Accept Parameters" in err

        scp.shutdown()

    def test_flag_log_collision(self):
        """Test error with -q -v and -d flag."""
        p = self.func(["-v", "-d"])
        p.wait()
        assert p.returncode != 0

    @pytest.mark.skip("No way to test comprehensively")
    def test_flag_log_level(self):
        """Test --log-level flag."""
        pass

    def test_flag_aet(self):
        """Test --calling-aet flag."""
        events = []

        def handle_find(event):
            events.append(event)
            yield 0x0000, None

        handlers = [
            (evt.EVT_C_FIND, handle_find),
        ]

        self.ae = ae = AE()
        ae.acse_timeout = 5
        ae.dimse_timeout = 5
        ae.network_timeout = 5
        ae.supported_contexts = (
            QueryRetrievePresentationContexts
            + BasicWorklistManagementPresentationContexts
        )
        scp = ae.start_server(("localhost", 11112), block=False, evt_handlers=handlers)

        p = self.func(["-aet", "MYSCU", "-k", "PatientName="])
        p.wait()
        assert p.returncode == 0

        scp.shutdown()

        assert events[0].event == evt.EVT_C_FIND
        requestor = events[0].assoc.requestor
        assert "MYSCU" == requestor.ae_title

    def test_flag_aec(self):
        """Test --called-aet flag."""
        events = []

        def handle_find(event):
            events.append(event)
            yield 0x0000, None

        handlers = [
            (evt.EVT_C_FIND, handle_find),
        ]

        self.ae = ae = AE()
        ae.acse_timeout = 5
        ae.dimse_timeout = 5
        ae.network_timeout = 5
        ae.supported_contexts = (
            QueryRetrievePresentationContexts
            + BasicWorklistManagementPresentationContexts
        )
        scp = ae.start_server(("localhost", 11112), block=False, evt_handlers=handlers)

        p = self.func(["-aec", "YOURSCP", "-k", "PatientName="])
        p.wait()
        assert p.returncode == 0

        scp.shutdown()

        assert events[0].event == evt.EVT_C_FIND
        requestor = events[0].assoc.requestor
        assert "YOURSCP" == requestor.primitive.called_ae_title

    def test_flag_ta(self, capfd):
        """Test --acse-timeout flag."""
        events = []

        def handle_requested(event):
            events.append(event)
            time.sleep(0.1)

        def handle_find(event):
            events.append(event)
            yield 0x0000, None

        def handle_abort(event):
            events.append(event)

        handlers = [
            (evt.EVT_C_FIND, handle_find),
            (evt.EVT_ABORTED, handle_abort),
            (evt.EVT_REQUESTED, handle_requested),
        ]

        self.ae = ae = AE()
        ae.acse_timeout = 5
        ae.dimse_timeout = 5
        ae.network_timeout = 5
        ae.supported_contexts = (
            QueryRetrievePresentationContexts
            + BasicWorklistManagementPresentationContexts
        )
        scp = ae.start_server(("localhost", 11112), block=False, evt_handlers=handlers)

        p = self.func(["-ta", "0.05", "-d", "-k", "PatientName="])
        p.wait()
        assert p.returncode == 1

        time.sleep(0.1)

        scp.shutdown()

        out, err = capfd.readouterr()
        assert "ACSE timeout reached while waiting for response" in err
        assert events[0].event == evt.EVT_REQUESTED
        assert events[1].event == evt.EVT_ABORTED

    def test_flag_td(self, capfd):
        """Test --dimse-timeout flag."""
        events = []

        def handle_find(event):
            events.append(event)
            time.sleep(0.1)
            yield 0x0000, None

        def handle_abort(event):
            events.append(event)

        handlers = [
            (evt.EVT_C_FIND, handle_find),
            (evt.EVT_ABORTED, handle_abort),
        ]

        self.ae = ae = AE()
        ae.acse_timeout = 5
        ae.dimse_timeout = 5
        ae.network_timeout = 5
        ae.supported_contexts = (
            QueryRetrievePresentationContexts
            + BasicWorklistManagementPresentationContexts
        )
        scp = ae.start_server(("localhost", 11112), block=False, evt_handlers=handlers)

        p = self.func(["-td", "0.05", "-d", "-k", "PatientName="])
        p.wait()
        assert p.returncode == 0

        time.sleep(0.1)

        scp.shutdown()

        out, err = capfd.readouterr()
        assert "DIMSE timeout reached while waiting for message" in err
        assert events[0].event == evt.EVT_C_FIND
        assert events[1].event == evt.EVT_ABORTED

    @pytest.mark.skip("Don't think this can be tested")
    def test_flag_tn(self, capfd):
        """Test --network-timeout flag."""
        pass

    def test_flag_max_pdu(self):
        """Test --max-pdu flag."""
        events = []

        def handle_find(event):
            events.append(event)
            yield 0x0000, None

        def handle_release(event):
            events.append(event)

        handlers = [(evt.EVT_C_FIND, handle_find), (evt.EVT_RELEASED, handle_release)]

        self.ae = ae = AE()
        ae.acse_timeout = 5
        ae.dimse_timeout = 5
        ae.network_timeout = 5
        ae.supported_contexts = (
            QueryRetrievePresentationContexts
            + BasicWorklistManagementPresentationContexts
        )
        scp = ae.start_server(("localhost", 11112), block=False, evt_handlers=handlers)

        p = self.func(["--max-pdu", "123456", "-k", "PatientName="])
        p.wait()
        assert p.returncode == 0

        scp.shutdown()

        assert events[0].event == evt.EVT_C_FIND
        assert events[1].event == evt.EVT_RELEASED
        requestor = events[1].assoc.requestor
        assert 123456 == requestor.maximum_length

    def test_flag_patient(self):
        """Test the -P flag."""
        events = []

        def handle_find(event):
            events.append(event)
            yield 0x0000, None

        handlers = [
            (evt.EVT_C_FIND, handle_find),
        ]

        self.ae = ae = AE()
        ae.acse_timeout = 5
        ae.dimse_timeout = 5
        ae.network_timeout = 5
        ae.supported_contexts = (
            QueryRetrievePresentationContexts
            + BasicWorklistManagementPresentationContexts
        )
        scp = ae.start_server(("localhost", 11112), block=False, evt_handlers=handlers)

        p = self.func(["-P", "-k", "PatientName="])
        p.wait()
        assert p.returncode == 0

        scp.shutdown()

        assert events[0].event == evt.EVT_C_FIND
        cx = events[0].context
        assert cx.abstract_syntax == (PatientRootQueryRetrieveInformationModelFind)

    def test_flag_study(self):
        """Test the -S flag."""
        events = []

        def handle_find(event):
            events.append(event)
            yield 0x0000, None

        handlers = [
            (evt.EVT_C_FIND, handle_find),
        ]

        self.ae = ae = AE()
        ae.acse_timeout = 5
        ae.dimse_timeout = 5
        ae.network_timeout = 5
        ae.supported_contexts = (
            QueryRetrievePresentationContexts
            + BasicWorklistManagementPresentationContexts
        )
        scp = ae.start_server(("localhost", 11112), block=False, evt_handlers=handlers)

        p = self.func(["-S", "-k", "PatientName="])
        p.wait()
        assert p.returncode == 0

        scp.shutdown()

        assert events[0].event == evt.EVT_C_FIND
        cx = events[0].context
        assert cx.abstract_syntax == StudyRootQueryRetrieveInformationModelFind

    def test_flag_patient_study(self):
        """Test the -O flag."""
        events = []

        def handle_find(event):
            events.append(event)
            yield 0x0000, None

        handlers = [
            (evt.EVT_C_FIND, handle_find),
        ]

        self.ae = ae = AE()
        ae.acse_timeout = 5
        ae.dimse_timeout = 5
        ae.network_timeout = 5
        ae.supported_contexts = (
            QueryRetrievePresentationContexts
            + BasicWorklistManagementPresentationContexts
        )
        scp = ae.start_server(("localhost", 11112), block=False, evt_handlers=handlers)

        p = self.func(["-O", "-k", "PatientName="])
        p.wait()
        assert p.returncode == 0

        scp.shutdown()

        assert events[0].event == evt.EVT_C_FIND
        cx = events[0].context
        assert cx.abstract_syntax == (PatientStudyOnlyQueryRetrieveInformationModelFind)

    def test_flag_ups(self):
        """Test the -U flag."""
        events = []

        def handle_find(event):
            events.append(event)
            yield 0x0000, None

        handlers = [
            (evt.EVT_C_FIND, handle_find),
        ]

        self.ae = ae = AE()
        ae.acse_timeout = 5
        ae.dimse_timeout = 5
        ae.network_timeout = 5
        ae.supported_contexts = (
            QueryRetrievePresentationContexts
            + BasicWorklistManagementPresentationContexts
            + UnifiedProcedurePresentationContexts
        )
        scp = ae.start_server(("localhost", 11112), block=False, evt_handlers=handlers)

        p = self.func(["-U", "-k", "PatientName="])
        p.wait()
        assert p.returncode == 0

        scp.shutdown()

        assert events[0].event == evt.EVT_C_FIND
        cx = events[0].context
        assert cx.abstract_syntax == UnifiedProcedureStepPull

    def test_flag_worklist(self):
        """Test the -W flag."""
        events = []

        def handle_find(event):
            events.append(event)
            yield 0x0000, None

        handlers = [
            (evt.EVT_C_FIND, handle_find),
        ]

        self.ae = ae = AE()
        ae.acse_timeout = 5
        ae.dimse_timeout = 5
        ae.network_timeout = 5
        ae.supported_contexts = (
            QueryRetrievePresentationContexts
            + BasicWorklistManagementPresentationContexts
        )
        scp = ae.start_server(("localhost", 11112), block=False, evt_handlers=handlers)

        p = self.func(["-W", "-k", "PatientName="])
        p.wait()
        assert p.returncode == 0

        scp.shutdown()

        assert events[0].event == evt.EVT_C_FIND
        cx = events[0].context
        assert cx.abstract_syntax == ModalityWorklistInformationFind

    def test_flag_write(self):
        """Test the -w flag."""

        def handle_find(event):
            yield 0xFF00, event.identifier

        handlers = [
            (evt.EVT_C_FIND, handle_find),
        ]

        self.ae = ae = AE()
        ae.acse_timeout = 5
        ae.dimse_timeout = 5
        ae.network_timeout = 5
        ae.supported_contexts = (
            QueryRetrievePresentationContexts
            + BasicWorklistManagementPresentationContexts
        )
        scp = ae.start_server(("localhost", 11112), block=False, evt_handlers=handlers)

        p = self.func(["-w", "-k", "PatientName="])
        p.wait()
        assert p.returncode == 0

        scp.shutdown()

        assert "rsp000001.dcm" in os.listdir()
        ds = dcmread("rsp000001.dcm")
        assert ds.PatientName == ""
        os.remove("rsp000001.dcm")


class TestFindSCU(FindSCUBase):
    """Tests for findscu.py"""

    def setup_method(self):
        """Run prior to each test"""
        self.ae = None
        self.func = start_findscu


class TestFindSCUCLI(FindSCUBase):
    """Tests for findscu.py"""

    def setup_method(self):
        """Run prior to each test"""
        self.ae = None
        self.func = start_findscu_scli



"""Association testing for DIMSE-N services"""

import queue
import time

import pytest

from pydicom.dataset import Dataset
from pydicom.tag import Tag
from pydicom.uid import ExplicitVRLittleEndian

from pynetdicom import AE, debug_logger, evt
from pynetdicom.sop_class import (
    DisplaySystem,
    Verification,
    PrintJob,
    ModalityPerformedProcedureStepNotification,
    ModalityPerformedProcedureStepRetrieve,
    ModalityPerformedProcedureStep,
    ProceduralEventLogging,
    BasicFilmSession,
    BasicGrayscalePrintManagementMeta,
    BasicColorPrintManagementMeta,
    Printer,
)


# debug_logger()


class DummyDIMSE:
    def __init__(self):
        self.status = None

    def send_msg(self, req, context_id):
        self.req = req
        self.context_id = context_id

    def get_msg(self, block=False):
        return None, None


class TestAssociationSendNEventReport:
    """Run tests on Association send_n_event_report."""

    def setup_method(self):
        self.ae = None

    def teardown_method(self):
        """Clear any active threads"""
        if self.ae:
            self.ae.shutdown()

    def test_must_be_associated(self):
        """Test can't send without association."""

        def handle(event):
            return 0x0000, Dataset()

        self.ae = ae = AE()
        ae.acse_timeout = 5
        ae.dimse_timeout = 5
        ae.network_timeout = 5
        ae.add_supported_context(ModalityPerformedProcedureStepNotification)
        scp = ae.start_server(
            ("localhost", 11112),
            block=False,
            evt_handlers=[(evt.EVT_N_EVENT_REPORT, handle)],
        )

        ae.add_requested_context(ModalityPerformedProcedureStepNotification)
        assoc = ae.associate("localhost", 11112)
        assert assoc.is_established
        assoc.release()
        assert not assoc.is_established

        with pytest.raises(RuntimeError):
            assoc.send_n_event_report(None, None, None, None)

        scp.shutdown()

    def test_no_abstract_syntax_match(self):
        """Test SCU when no accepted abstract syntax"""

        def handle(event):
            return 0x0000, Dataset()

        self.ae = ae = AE()
        ae.acse_timeout = 5
        ae.dimse_timeout = 5
        ae.network_timeout = 5
        ae.add_supported_context(ModalityPerformedProcedureStepNotification)
        scp = ae.start_server(
            ("localhost", 11112),
            block=False,
            evt_handlers=[(evt.EVT_N_EVENT_REPORT, handle)],
        )

        ae.add_requested_context(ModalityPerformedProcedureStepNotification)
        assoc = ae.associate("localhost", 11112)
        assert assoc.is_established

        msg = (
            r"No presentation context for 'Verification SOP Class' has been "
            r"accepted by the peer for the SCU role"
        )
        with pytest.raises(ValueError, match=msg):
            assoc.send_n_event_report(None, None, Verification, None)

        assoc.release()
        assert assoc.is_released

        scp.shutdown()

    def test_rq_bad_dataset_raises(self):
        """Test sending bad dataset raises exception."""

        def handle(event):
            return 0x0000, Dataset()

        self.ae = ae = AE()
        ae.acse_timeout = 5
        ae.dimse_timeout = 5
        ae.network_timeout = 5
        ae.add_supported_context(ModalityPerformedProcedureStepNotification)
        scp = ae.start_server(
            ("localhost", 11112),
            block=False,
            evt_handlers=[(evt.EVT_N_EVENT_REPORT, handle)],
        )

        ae.add_requested_context(
            ModalityPerformedProcedureStepNotification, ExplicitVRLittleEndian
        )
        assoc = ae.associate("localhost", 11112)
        assert assoc.is_established

        ds = Dataset()
        ds.PerimeterValue = b"\x00\x01"
        msg = r"Unable to encode the supplied 'Event Information' dataset"
        with pytest.raises(ValueError, match=msg):
            assoc.send_n_event_report(
                ds, 1, ModalityPerformedProcedureStepNotification, "1.2.3"
            )

        assoc.release()
        assert assoc.is_released

        scp.shutdown()

    def test_rsp_none(self):
        """Test no response from peer"""

        def handle(event):
            time.sleep(5)
            return 0x0000, Dataset()

        self.ae = ae = AE()
        ae.acse_timeout = 5
        ae.dimse_timeout = 0.2
        ae.network_timeout = 5
        ae.add_supported_context(ModalityPerformedProcedureStepNotification)
        scp = ae.start_server(
            ("localhost", 11112),
            block=False,
            evt_handlers=[(evt.EVT_N_EVENT_REPORT, handle)],
        )

        ae.add_requested_context(ModalityPerformedProcedureStepNotification)
        assoc = ae.associate("localhost", 11112)
        assert assoc.is_established

        ds = Dataset()
        ds.PatientName = "Test^test"
        status, ds = assoc.send_n_event_report(
            ds,
            1,
            ModalityPerformedProcedureStepNotification,
            "1.2.840.10008.5.1.1.40.1",
        )
        assert status == Dataset()
        assert ds is None
        assert assoc.is_aborted

        scp.shutdown()

    def test_rsp_invalid(self):
        """Test invalid DIMSE message received from peer"""

        def handle(event):
            return 0x0000, Dataset()

        self.ae = ae = AE()
        ae.acse_timeout = 5
        ae.dimse_timeout = 5
        ae.network_timeout = 5
        ae.add_supported_context(ModalityPerformedProcedureStepNotification)
        scp = ae.start_server(
            ("localhost", 11112),
            block=False,
            evt_handlers=[(evt.EVT_N_EVENT_REPORT, handle)],
        )

        ae.add_requested_context(ModalityPerformedProcedureStepNotification)
        assoc = ae.associate("localhost", 11112)
        assert assoc.is_established

        class DummyResponse:
            is_valid_response = False

        class DummyDIMSE:
            msg_queue = queue.Queue()
            gotten = False

            def send_msg(*args, **kwargs):
                return

            def get_msg(self, *args, **kwargs):
                if not self.gotten:
                    self.gotten = True
                    return None, DummyResponse()
                return None, None

        assoc._reactor_checkpoint.clear()
        while not assoc._is_paused:
            time.sleep(0.01)
        assoc.dimse = DummyDIMSE()

        ds = Dataset()
        ds.PatientName = "Test^test"
        status, ds = assoc.send_n_event_report(
            ds,
            1,
            ModalityPerformedProcedureStepNotification,
            "1.2.840.10008.5.1.1.40.1",
        )

        assert status == Dataset()
        assert ds is None
        assert assoc.is_aborted

        scp.shutdown()

    def test_rsp_failure(self):
        """Test receiving a failure response from the peer"""

        def handle(event):
            return 0x0112, None

        self.ae = ae = AE()
        ae.acse_timeout = 5
        ae.dimse_timeout = 5
        ae.network_timeout = 5
        ae.add_supported_context(ModalityPerformedProcedureStepNotification)
        scp = ae.start_server(
            ("localhost", 11112),
            block=False,
            evt_handlers=[(evt.EVT_N_EVENT_REPORT, handle)],
        )

        ae.add_requested_context(ModalityPerformedProcedureStepNotification)
        assoc = ae.associate("localhost", 11112)
        assert assoc.is_established

        ds = Dataset()
        ds.PatientName = "Test^test"
        status, ds = assoc.send_n_event_report(
            ds,
            1,
            ModalityPerformedProcedureStepNotification,
            "1.2.840.10008.5.1.1.40.1",
        )
        assert status.Status == 0x0112
        assert ds is None
        assoc.release()
        assert assoc.is_released

        scp.shutdown()

    def test_rsp_warning(self):
        """Test receiving a warning response from the peer"""

        def handle(event):
            return 0x0116, event.event_information

        self.ae = ae = AE()
        ae.acse_timeout = 5
        ae.dimse_timeout = 5
        ae.network_timeout = 5
        ae.add_supported_context(ModalityPerformedProcedureStepNotification)
        scp = ae.start_server(
            ("localhost", 11112),
            block=False,
            evt_handlers=[(evt.EVT_N_EVENT_REPORT, handle)],
        )

        ae.add_requested_context(ModalityPerformedProcedureStepNotification)
        assoc = ae.associate("localhost", 11112)
        assert assoc.is_established

        ds = Dataset()
        ds.PatientName = "Test^test"
        status, ds = assoc.send_n_event_report(
            ds,
            1,
            ModalityPerformedProcedureStepNotification,
            "1.2.840.10008.5.1.1.40.1",
        )
        assert status.Status == 0x0116
        assert ds.PatientName == "Test^test"
        assoc.release()
        assert assoc.is_released

        scp.shutdown()

    def test_rsp_success(self):
        """Test receiving a success response from the peer"""

        def handle(event):
            return 0x0000, event.event_information

        self.ae = ae = AE()
        ae.acse_timeout = 5
        ae.dimse_timeout = 5
        ae.network_timeout = 5
        ae.add_supported_context(ModalityPerformedProcedureStepNotification)
        scp = ae.start_server(
            ("localhost", 11112),
            block=False,
            evt_handlers=[(evt.EVT_N_EVENT_REPORT, handle)],
        )

        ae.add_requested_context(ModalityPerformedProcedureStepNotification)
        assoc = ae.associate("localhost", 11112)
        assert assoc.is_established

        ds = Dataset()
        ds.PatientName = "Test^test"
        status, ds = assoc.send_n_event_report(
            ds,
            1,
            ModalityPerformedProcedureStepNotification,
            "1.2.840.10008.5.1.1.40.1",
        )
        assert status.Status == 0x0000
        assert ds.PatientName == "Test^test"
        assoc.release()
        assert assoc.is_released

        scp.shutdown()

    def test_rsp_unknown_status(self):
        """Test unknown status value returned by peer"""

        def handle(event):
            return 0xFFF0, event.event_information

        self.ae = ae = AE()
        ae.acse_timeout = 5
        ae.dimse_timeout = 5
        ae.network_timeout = 5
        ae.add_supported_context(ModalityPerformedProcedureStepNotification)
        scp = ae.start_server(
            ("localhost", 11112),
            block=False,
            evt_handlers=[(evt.EVT_N_EVENT_REPORT, handle)],
        )

        ae.add_requested_context(ModalityPerformedProcedureStepNotification)
        assoc = ae.associate("localhost", 11112)
        assert assoc.is_established

        ds = Dataset()
        ds.PatientName = "Test^test"
        status, ds = assoc.send_n_event_report(
            ds,
            1,
            ModalityPerformedProcedureStepNotification,
            "1.2.840.10008.5.1.1.40.1",
        )
        assert status.Status == 0xFFF0
        assert ds is None
        assoc.release()
        assert assoc.is_released

        scp.shutdown()

    def test_rsp_bad_dataset(self):
        """Test handler returns bad dataset"""

        def handle(event):
            def test():
                pass

            return 0x0000, test

        self.ae = ae = AE()
        ae.add_requested_context(ModalityPerformedProcedureStepNotification)
        ae.add_supported_context(ModalityPerformedProcedureStepNotification)

        handlers = [(evt.EVT_N_EVENT_REPORT, handle)]
        scp = ae.start_server(("localhost", 11112), evt_handlers=handlers, block=False)

        ae.acse_timeout = 5
        ae.dimse_timeout = 5
        assoc = ae.associate("localhost", 11112)

        assert assoc.is_established

        # Event Information
        ds = Dataset()
        ds.PatientName = "Test^test"
        status, ds = assoc.send_n_event_report(
            ds,
            1,
            ModalityPerformedProcedureStepNotification,
            "1.2.840.10008.5.1.1.40.1",
        )

        assert status.Status == 0x0110
        assert ds is None

        assoc.release()
        scp.shutdown()

    def test_decode_failure(self):
        """Test being unable to decode received dataset"""

        def handle(event):
            def test():
                pass

            return 0x0000, test

        self.ae = ae = AE()
        ae.add_requested_context(
            ModalityPerformedProcedureStepNotification, ExplicitVRLittleEndian
        )
        ae.add_supported_context(ModalityPerformedProcedureStepNotification)

        handlers = [(evt.EVT_N_EVENT_REPORT, handle)]
        scp = ae.start_server(("localhost", 11112), evt_handlers=handlers, block=False)

        ae.acse_timeout = 5
        ae.dimse_timeout = 5
        assoc = ae.associate("localhost", 11112)

        class DummyReply:
            def getvalue(self):
                def test():
                    pass

                return test

        class DummyMessage:
            is_valid_response = True
            EventReply = DummyReply()
            Status = 0x0000
            STATUS_OPTIONAL_KEYWORDS = []

        class DummyDIMSE:
            msg_queue = queue.Queue()
            gotten = False

            def send_msg(*args, **kwargs):
                return

            def get_msg(self, *args, **kwargs):
                if not self.gotten:
                    self.gotten = True
                    return 1, DummyMessage()
                return None, None

        assoc._reactor_checkpoint.clear()
        while not assoc._is_paused:
            time.sleep(0.01)
        assoc.dimse = DummyDIMSE()
        assert assoc.is_established

        # Event Information
        ds = Dataset()
        ds.PatientName = "Test^test"
        status, ds = assoc.send_n_event_report(
            ds,
            1,
            ModalityPerformedProcedureStepNotification,
            "1.2.840.10008.5.1.1.40.1",
        )

        assert status.Status == 0x0110
        assert ds is None

        assoc.release()
        scp.shutdown()

    def test_extra_status(self):
        """Test extra status elements are available."""

        def handle(event):
            status = Dataset()
            status.Status = 0xFFF0
            status.ErrorComment = "Some comment"
            status.ErrorID = 12
            status.AffectedSOPClassUID = "1.2.3"
            status.AffectedSOPInstanceUID = "1.2.3.4"
            return status, event.event_information

        self.ae = ae = AE()
        ae.acse_timeout = 5
        ae.dimse_timeout = 5
        ae.network_timeout = 5
        ae.add_supported_context(ModalityPerformedProcedureStepNotification)
        scp = ae.start_server(
            ("localhost", 11112),
            block=False,
            evt_handlers=[(evt.EVT_N_EVENT_REPORT, handle)],
        )

        ae.add_requested_context(ModalityPerformedProcedureStepNotification)
        assoc = ae.associate("localhost", 11112)
        assert assoc.is_established

        ds = Dataset()
        ds.PatientName = "Test^test"
        status, ds = assoc.send_n_event_report(
            ds,
            1,
            ModalityPerformedProcedureStepNotification,
            "1.2.840.10008.5.1.1.40.1",
        )
        assert status.Status == 0xFFF0
        assert status.ErrorComment == "Some comment"
        assert status.ErrorID == 12
        assert status.AffectedSOPClassUID == "1.2.3"
        assert status.AffectedSOPInstanceUID == "1.2.3.4"
        assert ds is None
        assoc.release()
        assert assoc.is_released

        scp.shutdown()

    def test_meta_uid(self):
        """Test using a Meta SOP Class"""
        self.ae = ae = AE()
        ae.acse_timeout = 5
        ae.dimse_timeout = 5
        ae.network_timeout = 5
        ae.add_supported_context(BasicGrayscalePrintManagementMeta)
        ae.add_supported_context(Printer)
        scp = ae.start_server(("localhost", 11112), block=False)

        ae.add_requested_context(BasicGrayscalePrintManagementMeta)
        ae.add_requested_context(Printer)
        assoc = ae.associate("localhost", 11112)
        assert assoc.is_established

        assoc.dimse = DummyDIMSE()

        ds = Dataset()
        ds.PatientName = "Test^test"
        # Receives None, None from DummyDIMSE, aborts
        status, ds = assoc.send_n_event_report(
            ds,
            1,
            Printer,
            "1.2.840.10008.5.1.1.40.1",
            meta_uid=BasicGrayscalePrintManagementMeta,
        )
        assert assoc.is_aborted

        scp.shutdown()

        assert assoc.dimse.req.AffectedSOPClassUID == Printer
        assert assoc.dimse.context_id == 1
        assert (
            assoc._accepted_cx[1].abstract_syntax == BasicGrayscalePrintManagementMeta
        )

    def test_meta_uid_good(self):
        """Test sending a request using a Meta SOP Class."""
        handler_data = []

        def handle(event):
            handler_data.append(event)
            return 0x0000, event.event_information

        self.ae = ae = AE()
        ae.acse_timeout = 5
        ae.dimse_timeout = 5
        ae.network_timeout = 5
        ae.add_supported_context(BasicGrayscalePrintManagementMeta)
        scp = ae.start_server(
            ("localhost", 11112),
            block=False,
            evt_handlers=[(evt.EVT_N_EVENT_REPORT, handle)],
        )

        ae.add_requested_context(BasicGrayscalePrintManagementMeta)
        assoc = ae.associate("localhost", 11112)
        assert assoc.is_established

        ds = Dataset()
        ds.PatientName = "Test^test"
        status, ds = assoc.send_n_event_report(
            ds,
            1,
            BasicFilmSession,
            "1.2.840.10008.5.1.1.40.1",
            meta_uid=BasicGrayscalePrintManagementMeta,
        )
        assert status.Status == 0x0000
        assert ds.PatientName == "Test^test"
        assoc.release()
        assert assoc.is_released

        scp.shutdown()

        req = handler_data[0].request
        cx = handler_data[0].context

        assert req.AffectedSOPClassUID == BasicFilmSession
        assert cx.abstract_syntax == BasicGrayscalePrintManagementMeta

    def test_meta_uid_bad(self):
        """Test sending a request using a Meta SOP Class."""

        def handle(event):
            return 0x0000, event.event_information

        self.ae = ae = AE()
        ae.acse_timeout = 5
        ae.dimse_timeout = 5
        ae.network_timeout = 5
        ae.add_supported_context(BasicGrayscalePrintManagementMeta)
        scp = ae.start_server(
            ("localhost", 11112),
            block=False,
            evt_handlers=[(evt.EVT_N_EVENT_REPORT, handle)],
        )

        ae.add_requested_context(BasicGrayscalePrintManagementMeta)
        assoc = ae.associate("localhost", 11112)
        assert assoc.is_established

        ds = Dataset()
        ds.PatientName = "Test^test"
        msg = (
            r"No presentation context for 'Basic Color Print Management "
            r"Meta SOP Class' has been "
            r"accepted by the peer "
            r"for the SCU role"
        )
        with pytest.raises(ValueError, match=msg):
            assoc.send_n_event_report(
                ds,
                1,
                BasicFilmSession,
                "1.2.840.10008.5.1.1.40.1",
                meta_uid=BasicColorPrintManagementMeta,
            )

        assoc.release()
        assert assoc.is_released

        scp.shutdown()


class TestAssociationSendNGet:
    """Run tests on Association send_n_get."""

    def setup_method(self):
        """Run prior to each test"""
        self.ae = None

    def teardown_method(self):
        """Clear any active threads"""
        if self.ae:
            self.ae.shutdown()

    def test_must_be_associated(self):
        """Test can't send without association."""

        def handle(event):
            ds = Dataset()
            ds.PatientName = "Test^test"
            return 0x0000, ds

        self.ae = ae = AE()
        ae.acse_timeout = 5
        ae.dimse_timeout = 5
        ae.network_timeout = 5
        ae.add_supported_context(DisplaySystem)
        scp = ae.start_server(
            ("localhost", 11112), block=False, evt_handlers=[(evt.EVT_N_GET, handle)]
        )

        ae.add_requested_context(DisplaySystem)
        assoc = ae.associate("localhost", 11112)
        assoc.release()
        assert assoc.is_released
        assert not assoc.is_established
        with pytest.raises(RuntimeError):
            assoc.send_n_get(None, None, None)

        scp.shutdown()

    def test_no_abstract_syntax_match(self):
        """Test SCU when no accepted abstract syntax"""

        def handle(event):
            ds = Dataset()
            ds.PatientName = "Test^test"
            return 0x0000, ds

        self.ae = ae = AE()
        ae.acse_timeout = 5
        ae.dimse_timeout = 5
        ae.network_timeout = 5
        ae.add_supported_context(DisplaySystem)
        scp = ae.start_server(
            ("localhost", 11112), block=False, evt_handlers=[(evt.EVT_N_GET, handle)]
        )

        ae.add_requested_context(DisplaySystem)
        assoc = ae.associate("localhost", 11112)
        assert assoc.is_established
        msg = (
            r"No presentation context for 'Verification SOP Class' has been "
            r"accepted by the peer "
            r"for the SCU role"
        )
        with pytest.raises(ValueError, match=msg):
            assoc.send_n_get(None, Verification, None)

        assoc.release()
        assert assoc.is_released

        scp.shutdown()

    def test_rsp_none(self):
        """Test no response from peer"""

        def handle(event):
            ds = Dataset()
            ds.PatientName = "Test^test"
            time.sleep(5)
            return 0x0000, ds

        self.ae = ae = AE()
        ae.acse_timeout = 5
        ae.dimse_timeout = 0.1
        ae.network_timeout = 5
        ae.add_supported_context(DisplaySystem)
        scp = ae.start_server(
            ("localhost", 11112), block=False, evt_handlers=[(evt.EVT_N_GET, handle)]
        )

        ae.add_requested_context(DisplaySystem)
        assoc = ae.associate("localhost", 11112)
        assert assoc.is_established

        status, ds = assoc.send_n_get(
            [(0x7FE0, 0x0010)], DisplaySystem, "1.2.840.10008.5.1.1.40.1"
        )

        assert status == Dataset()
        assert ds is None
        assert assoc.is_aborted

        scp.shutdown()

    def test_rsp_invalid(self):
        """Test invalid DIMSE message received from peer"""

        def handle(event):
            ds = Dataset()
            ds.PatientName = "Test^test"
            return 0x0000, ds

        self.ae = ae = AE()
        ae.acse_timeout = 5
        ae.dimse_timeout = 5
        ae.network_timeout = 5
        ae.add_supported_context(DisplaySystem)
        scp = ae.start_server(
            ("localhost", 11112), block=False, evt_handlers=[(evt.EVT_N_GET, handle)]
        )

        ae.add_requested_context(DisplaySystem)
        assoc = ae.associate("localhost", 11112)
        assert assoc.is_established

        class DummyResponse:
            is_valid_response = False

        class DummyDIMSE:
            msg_queue = queue.Queue()
            gotten = False

            def send_msg(*args, **kwargs):
                return

            def get_msg(self, *args, **kwargs):
                if not self.gotten:
                    self.gotten = True
                    return None, DummyResponse()
                return None, None

        assoc._reactor_checkpoint.clear()
        while not assoc._is_paused:
            time.sleep(0.01)
        assoc.dimse = DummyDIMSE()
        assert assoc.is_established
        status, ds = assoc.send_n_get(
            [(0x7FE0, 0x0010)], DisplaySystem, "1.2.840.10008.5.1.1.40.1"
        )
        assert status == Dataset()
        assert ds is None
        assert assoc.is_aborted

        scp.shutdown()

    def test_rsp_failure(self):
        """Test receiving a failure response from the peer"""

        def handle(event):
            return 0x0112, None

        self.ae = ae = AE()
        ae.acse_timeout = 5
        ae.dimse_timeout = 5
        ae.network_timeout = 5
        ae.add_supported_context(DisplaySystem)
        scp = ae.start_server(
            ("localhost", 11112), block=False, evt_handlers=[(evt.EVT_N_GET, handle)]
        )

        ae.add_requested_context(DisplaySystem)
        assoc = ae.associate("localhost", 11112)
        assert assoc.is_established

        status, ds = assoc.send_n_get(
            [(0x7FE0, 0x0010)], DisplaySystem, "1.2.840.10008.5.1.1.40.1"
        )
        assert status.Status == 0x0112
        assert ds is None
        assoc.release()
        assert assoc.is_released

        scp.shutdown()

    def test_rsp_warning(self):
        """Test receiving a warning response from the peer"""

        def handle(event):
            ds = Dataset()
            ds.PatientName = "Test"
            ds.SOPClassUID = DisplaySystem
            ds.SOPInstanceUID = "1.2.3.4"
            return 0x0116, ds

        self.ae = ae = AE()
        ae.acse_timeout = 5
        ae.dimse_timeout = 5
        ae.network_timeout = 5
        ae.add_supported_context(DisplaySystem)
        scp = ae.start_server(
            ("localhost", 11112), block=False, evt_handlers=[(evt.EVT_N_GET, handle)]
        )

        ae.add_requested_context(DisplaySystem)
        assoc = ae.associate("localhost", 11112)
        assert assoc.is_established

        status, ds = assoc.send_n_get(
            [(0x7FE0, 0x0010)], DisplaySystem, "1.2.840.10008.5.1.1.40.1"
        )
        assert status.Status == 0x0116
        assert ds is not None
        assert isinstance(ds, Dataset)
        assert ds.PatientName == "Test"
        assert ds.SOPClassUID == DisplaySystem
        assert ds.SOPInstanceUID == "1.2.3.4"
        assoc.release()
        assert assoc.is_released

        scp.shutdown()

    def test_rsp_success(self):
        """Test receiving a success response from the peer"""

        def handle(event):
            ds = Dataset()
            ds.PatientName = "Test"
            ds.SOPClassUID = DisplaySystem
            ds.SOPInstanceUID = "1.2.3.4"
            return 0x0000, ds

        self.ae = ae = AE()
        ae.acse_timeout = 5
        ae.dimse_timeout = 5
        ae.network_timeout = 5
        ae.add_supported_context(DisplaySystem)
        scp = ae.start_server(
            ("localhost", 11112), block=False, evt_handlers=[(evt.EVT_N_GET, handle)]
        )

        ae.add_requested_context(DisplaySystem)
        assoc = ae.associate("localhost", 11112)
        assert assoc.is_established

        status, ds = assoc.send_n_get(
            [(0x7FE0, 0x0010)], DisplaySystem, "1.2.840.10008.5.1.1.40.1"
        )
        assert status.Status == 0x0000
        assert ds is not None
        assert isinstance(ds, Dataset)
        assert ds.PatientName == "Test"
        assert ds.SOPClassUID == DisplaySystem
        assert ds.SOPInstanceUID == "1.2.3.4"
        assoc.release()
        assert assoc.is_released

        scp.shutdown()

    def test_rsp_unknown_status(self):
        """Test unknown status value returned by peer"""

        def handle(event):
            ds = Dataset()
            ds.PatientName = "Test"
            ds.SOPClassUID = DisplaySystem
            ds.SOPInstanceUID = "1.2.3.4"
            return 0xFFF0, ds

        self.ae = ae = AE()
        ae.acse_timeout = 5
        ae.dimse_timeout = 5
        ae.network_timeout = 5
        ae.add_supported_context(DisplaySystem)
        scp = ae.start_server(
            ("localhost", 11112), block=False, evt_handlers=[(evt.EVT_N_GET, handle)]
        )

        ae.add_requested_context(DisplaySystem)
        assoc = ae.associate("localhost", 11112)
        assert assoc.is_established

        status, ds = assoc.send_n_get(
            [(0x7FE0, 0x0010)], DisplaySystem, "1.2.840.10008.5.1.1.40.1"
        )
        assert status.Status == 0xFFF0
        assert ds is None
        assoc.release()
        assert assoc.is_released

        scp.shutdown()

    def test_rsp_bad_dataset(self):
        """Test handler returns bad dataset"""

        def handle(event):
            def test():
                pass

            return 0x0000, test

        self.ae = ae = AE()
        ae.add_requested_context(ModalityPerformedProcedureStepRetrieve)
        ae.add_supported_context(ModalityPerformedProcedureStepRetrieve)

        handlers = [(evt.EVT_N_GET, handle)]
        scp = ae.start_server(("localhost", 11112), evt_handlers=handlers, block=False)

        ae.acse_timeout = 5
        ae.dimse_timeout = 5
        assoc = ae.associate("localhost", 11112)

        assert assoc.is_established

        # Event Information
        attrs = [0x00100010, 0x00100020]
        status, ds = assoc.send_n_get(
            attrs, ModalityPerformedProcedureStepRetrieve, "1.2.840.10008.5.1.1.40.1"
        )

        assert status.Status == 0x0110
        assert ds is None

        assoc.release()
        scp.shutdown()

    def test_decode_failure(self):
        """Test bad dataset received from peer"""

        def handle(event):
            ds = Dataset()
            ds.PatientName = "Test^test"
            return 0x0000, ds

        self.ae = ae = AE()
        ae.acse_timeout = 5
        ae.dimse_timeout = 5
        ae.network_timeout = 5
        ae.add_supported_context(DisplaySystem)
        scp = ae.start_server(
            ("localhost", 11112), block=False, evt_handlers=[(evt.EVT_N_GET, handle)]
        )

        ae.add_requested_context(DisplaySystem)
        assoc = ae.associate("localhost", 11112)
        assert assoc.is_established

        class DummyReply:
            def getvalue(self):
                def test():
                    pass

                return test

        class DummyMessage:
            is_valid_response = True
            AttributeList = DummyReply()
            Status = 0x0000
            STATUS_OPTIONAL_KEYWORDS = []

        class DummyDIMSE:
            msg_queue = queue.Queue()
            gotten = False

            def send_msg(*args, **kwargs):
                return

            def get_msg(self, *args, **kwargs):
                if not self.gotten:
                    self.gotten = True
                    return 1, DummyMessage()
                return None, None

        assoc._reactor_checkpoint.clear()
        while not assoc._is_paused:
            time.sleep(0.01)
        assoc.dimse = DummyDIMSE()
        assert assoc.is_established
        status, ds = assoc.send_n_get(
            [(0x7FE0, 0x0010)], DisplaySystem, "1.2.840.10008.5.1.1.40.1"
        )

        assert status.Status == 0x0110
        assert ds is None

        scp.shutdown()

    def test_extra_status(self):
        """Test extra status elements are available."""

        def handle(event):
            ds = Dataset()
            ds.Status = 0xFFF0
            ds.ErrorComment = "Some comment"
            ds.ErrorID = 12
            ds.AttributeIdentifierList = 0x00100020
            return ds, None

        self.ae = ae = AE()
        ae.acse_timeout = 5
        ae.dimse_timeout = 5
        ae.network_timeout = 5
        ae.add_supported_context(DisplaySystem)
        scp = ae.start_server(
            ("localhost", 11112), block=False, evt_handlers=[(evt.EVT_N_GET, handle)]
        )

        ae.add_requested_context(DisplaySystem)
        assoc = ae.associate("localhost", 11112)
        assert assoc.is_established

        status, ds = assoc.send_n_get(
            [(0x7FE0, 0x0010)], DisplaySystem, "1.2.840.10008.5.1.1.40.1"
        )
        assert status.Status == 0xFFF0
        assert status.ErrorComment == "Some comment"
        assert status.ErrorID == 12
        assert status.AttributeIdentifierList == 0x00100020
        assert ds is None
        assoc.release()
        assert assoc.is_released

        scp.shutdown()

    def test_meta_uid(self):
        """Test using a Meta SOP Class"""
        self.ae = ae = AE()
        ae.acse_timeout = 5
        ae.dimse_timeout = 5
        ae.network_timeout = 5
        ae.add_supported_context(BasicGrayscalePrintManagementMeta)
        ae.add_supported_context(Printer)
        scp = ae.start_server(("localhost", 11112), block=False)

        ae.add_requested_context(BasicGrayscalePrintManagementMeta)
        ae.add_requested_context(Printer)
        assoc = ae.associate("localhost", 11112)
        assert assoc.is_established

        assoc.dimse = DummyDIMSE()

        ds = Dataset()
        ds.PatientName = "Test^test"
        # Receives None, None from DummyDIMSE, aborts
        status, ds = assoc.send_n_get(
            [(0x00100010)],
            Printer,
            "1.2.840.10008.5.1.1.40.1",
            meta_uid=BasicGrayscalePrintManagementMeta,
        )
        assert assoc.is_aborted

        scp.shutdown()

        assert assoc.dimse.req.RequestedSOPClassUID == Printer
        assert assoc.dimse.context_id == 1
        assert (
            assoc._accepted_cx[1].abstract_syntax == BasicGrayscalePrintManagementMeta
        )

    def test_meta_uid_good(self):
        """Test sending a request using a Meta SOP Class."""
        handler_data = []

        def handle(event):
            handler_data.append(event)
            ds = Dataset()
            ds.PatientName = "Test"
            ds.SOPClassUID = DisplaySystem
            ds.SOPInstanceUID = "1.2.3.4"
            return 0x0000, ds

        self.ae = ae = AE()
        ae.acse_timeout = 5
        ae.dimse_timeout = 5
        ae.network_timeout = 5
        ae.add_supported_context(BasicGrayscalePrintManagementMeta)
        scp = ae.start_server(
            ("localhost", 11112), block=False, evt_handlers=[(evt.EVT_N_GET, handle)]
        )

        ae.add_requested_context(BasicGrayscalePrintManagementMeta)
        assoc = ae.associate("localhost", 11112)
        assert assoc.is_established

        ds = Dataset()
        ds.PatientName = "Test^test"
        status, ds = assoc.send_n_get(
            [(0x7FE0, 0x0010)],
            DisplaySystem,
            "1.2.840.10008.5.1.1.40.1",
            meta_uid=BasicGrayscalePrintManagementMeta,
        )

        assert status.Status == 0x0000
        assert ds is not None
        assert isinstance(ds, Dataset)
        assert ds.PatientName == "Test"
        assert ds.SOPClassUID == DisplaySystem
        assert ds.SOPInstanceUID == "1.2.3.4"
        assoc.release()
        assert assoc.is_released

        scp.shutdown()

        req = handler_data[0].request
        cx = handler_data[0].context

        assert req.RequestedSOPClassUID == DisplaySystem
        assert cx.abstract_syntax == BasicGrayscalePrintManagementMeta

    def test_meta_uid_bad(self):
        """Test sending a request using a Meta SOP Class."""

        def handle(event):
            ds = Dataset()
            ds.PatientName = "Test"
            ds.SOPClassUID = DisplaySystem
            ds.SOPInstanceUID = "1.2.3.4"
            return 0x0000, ds

        self.ae = ae = AE()
        ae.acse_timeout = 5
        ae.dimse_timeout = 5
        ae.network_timeout = 5
        ae.add_supported_context(BasicGrayscalePrintManagementMeta)
        scp = ae.start_server(
            ("localhost", 11112), block=False, evt_handlers=[(evt.EVT_N_GET, handle)]
        )

        ae.add_requested_context(BasicGrayscalePrintManagementMeta)
        assoc = ae.associate("localhost", 11112)
        assert assoc.is_established

        ds = Dataset()
        ds.PatientName = "Test^test"
        msg = (
            r"No presentation context for 'Basic Color Print Management "
            r"Meta SOP Class' has been "
            r"accepted by the peer "
            r"for the SCU role"
        )
        with pytest.raises(ValueError, match=msg):
            assoc.send_n_get(
                [(0x7FE0, 0x0010)],
                DisplaySystem,
                "1.2.840.10008.5.1.1.40.1",
                meta_uid=BasicColorPrintManagementMeta,
            )

        assoc.release()
        assert assoc.is_released

        scp.shutdown()


class TestAssociationSendNSet:
    """Run tests on Association send_n_set."""

    def setup_method(self):
        self.ae = None

    def teardown_method(self):
        """Clear any active threads"""
        if self.ae:
            self.ae.shutdown()

    def test_must_be_associated(self):
        """Test can't send without association."""

        def handle(event):
            ds = Dataset()
            ds.PatientName = "Test^test"
            return 0x0000, ds

        self.ae = ae = AE()
        ae.acse_timeout = 5
        ae.dimse_timeout = 5
        ae.network_timeout = 5
        ae.add_supported_context(ModalityPerformedProcedureStep)
        scp = ae.start_server(
            ("localhost", 11112), block=False, evt_handlers=[(evt.EVT_N_SET, handle)]
        )

        ae.add_requested_context(ModalityPerformedProcedureStep)
        assoc = ae.associate("localhost", 11112)
        assoc.release()
        assert assoc.is_released
        assert not assoc.is_established
        with pytest.raises(RuntimeError):
            assoc.send_n_set(None, None, None)

        scp.shutdown()

    def test_no_abstract_syntax_match(self):
        """Test SCU when no accepted abstract syntax"""

        def handle(event):
            ds = Dataset()
            ds.PatientName = "Test^test"
            return 0x0000, ds

        self.ae = ae = AE()
        ae.acse_timeout = 5
        ae.dimse_timeout = 5
        ae.network_timeout = 5
        ae.add_supported_context(ModalityPerformedProcedureStep)
        scp = ae.start_server(
            ("localhost", 11112), block=False, evt_handlers=[(evt.EVT_N_SET, handle)]
        )

        ae.add_requested_context(ModalityPerformedProcedureStep)
        assoc = ae.associate("localhost", 11112)
        assert assoc.is_established

        msg = (
            r"No presentation context for 'Verification SOP Class' has been "
            r"accepted by the peer "
            r"for the SCU role"
        )
        with pytest.raises(ValueError, match=msg):
            assoc.send_n_set(None, Verification, None)

        assoc.release()
        assert assoc.is_released

        scp.shutdown()

    def test_rq_bad_dataset_raises(self):
        """Test sending bad dataset raises exception."""

        def handle(event):
            ds = Dataset()
            ds.PatientName = "Test^test"
            return 0x0000, ds

        self.ae = ae = AE()
        ae.acse_timeout = 5
        ae.dimse_timeout = 5
        ae.network_timeout = 5
        ae.add_supported_context(ModalityPerformedProcedureStep)
        scp = ae.start_server(
            ("localhost", 11112), block=False, evt_handlers=[(evt.EVT_N_SET, handle)]
        )

        ae.add_requested_context(ModalityPerformedProcedureStep, ExplicitVRLittleEndian)
        assoc = ae.associate("localhost", 11112)
        assert assoc.is_established

        mod_list = Dataset()
        mod_list.PerimeterValue = b"\x00\x01"
        msg = r"Failed to encode the supplied 'Modification List' dataset"
        with pytest.raises(ValueError, match=msg):
            assoc.send_n_set(mod_list, ModalityPerformedProcedureStep, "1.2.3")

        assoc.release()
        assert assoc.is_released

        scp.shutdown()

    def test_rsp_none(self):
        """Test no response from peer"""

        def handle(event):
            ds = Dataset()
            ds.PatientName = "Test^test"
            time.sleep(5)
            return 0x0000, ds

        self.ae = ae = AE()
        ae.acse_timeout = 5
        ae.dimse_timeout = 0.2
        ae.network_timeout = 5
        ae.add_supported_context(ModalityPerformedProcedureStep)
        scp = ae.start_server(
            ("localhost", 11112), block=False, evt_handlers=[(evt.EVT_N_SET, handle)]
        )

        ae.add_requested_context(ModalityPerformedProcedureStep)
        assoc = ae.associate("localhost", 11112)
        assert assoc.is_established

        mod_list = Dataset()
        mod_list.PatientName = "Test^test"
        status, ds = assoc.send_n_set(
            mod_list, ModalityPerformedProcedureStep, "1.2.840.10008.5.1.1.40.1"
        )

        assert status == Dataset()
        assert ds is None
        assert assoc.is_aborted

        scp.shutdown()

    def test_rsp_invalid(self):
        """Test invalid DIMSE message received from peer"""

        def handle(event):
            ds = Dataset()
            ds.PatientName = "Test^test"
            return 0x0000, ds

        self.ae = ae = AE()
        ae.acse_timeout = 5
        ae.dimse_timeout = 0.4
        ae.network_timeout = 5
        ae.add_supported_context(ModalityPerformedProcedureStep)
        scp = ae.start_server(
            ("localhost", 11112), block=False, evt_handlers=[(evt.EVT_N_SET, handle)]
        )

        ae.add_requested_context(ModalityPerformedProcedureStep)
        assoc = ae.associate("localhost", 11112)
        assert assoc.is_established

        class DummyResponse:
            is_valid_response = False

        class DummyDIMSE:
            msg_queue = queue.Queue()
            gotten = False

            def send_msg(*args, **kwargs):
                return

            def get_msg(self, *args, **kwargs):
                if not self.gotten:
                    self.gotten = True
                    return None, DummyResponse()
                return None, None

        assoc._reactor_checkpoint.clear()
        while not assoc._is_paused:
            time.sleep(0.01)
        assoc.dimse = DummyDIMSE()
        mod_list = Dataset()
        mod_list.PatientName = "Test^test"
        status, ds = assoc.send_n_set(
            mod_list, ModalityPerformedProcedureStep, "1.2.840.10008.5.1.1.40.1"
        )

        assert status == Dataset()
        assert ds is None
        assert assoc.is_aborted

        scp.shutdown()

    def test_rsp_failure(self):
        """Test receiving a failure response from the peer"""

        def handle(event):
            return 0x0112, None

        self.ae = ae = AE()
        ae.acse_timeout = 5
        ae.dimse_timeout = 5
        ae.network_timeout = 5
        ae.add_supported_context(ModalityPerformedProcedureStep)
        scp = ae.start_server(
            ("localhost", 11112), block=False, evt_handlers=[(evt.EVT_N_SET, handle)]
        )

        ae.add_requested_context(ModalityPerformedProcedureStep)
        assoc = ae.associate("localhost", 11112)
        assert assoc.is_established

        ds = Dataset()
        ds.PatientName = "Test^test"
        status, ds = assoc.send_n_set(
            ds, ModalityPerformedProcedureStep, "1.2.840.10008.5.1.1.40.1"
        )
        assert status.Status == 0x0112
        assert ds is None
        assoc.release()
        assert assoc.is_released

        scp.shutdown()

    def test_rsp_warning(self):
        """Test receiving a warning response from the peer"""

        def handle(event):
            return 0x0116, event.modification_list

        self.ae = ae = AE()
        ae.acse_timeout = 5
        ae.dimse_timeout = 5
        ae.network_timeout = 5
        ae.add_supported_context(ModalityPerformedProcedureStep)
        scp = ae.start_server(
            ("localhost", 11112), block=False, evt_handlers=[(evt.EVT_N_SET, handle)]
        )

        ae.add_requested_context(ModalityPerformedProcedureStep)
        assoc = ae.associate("localhost", 11112)
        assert assoc.is_established

        ds = Dataset()
        ds.PatientName = "Test^test"
        status, ds = assoc.send_n_set(
            ds, ModalityPerformedProcedureStep, "1.2.840.10008.5.1.1.40.1"
        )
        assert status.Status == 0x0116
        assert ds.PatientName == "Test^test"
        assoc.release()
        assert assoc.is_released

        scp.shutdown()

    def test_rsp_success(self):
        """Test receiving a success response from the peer"""

        def handle(event):
            return 0x0000, event.modification_list

        self.ae = ae = AE()
        ae.acse_timeout = 5
        ae.dimse_timeout = 5
        ae.network_timeout = 5
        ae.add_supported_context(ModalityPerformedProcedureStep)
        scp = ae.start_server(
            ("localhost", 11112), block=False, evt_handlers=[(evt.EVT_N_SET, handle)]
        )

        ae.add_requested_context(ModalityPerformedProcedureStep)
        assoc = ae.associate("localhost", 11112)
        assert assoc.is_established

        ds = Dataset()
        ds.PatientName = "Test^test"
        status, ds = assoc.send_n_set(
            ds, ModalityPerformedProcedureStep, "1.2.840.10008.5.1.1.40.1"
        )
        assert status.Status == 0x0000
        assert ds.PatientName == "Test^test"
        assoc.release()
        assert assoc.is_released

        scp.shutdown()

    def test_rsp_unknown_status(self):
        """Test unknown status value returned by peer"""

        def handle(event):
            return 0xFFF0, event.modification_list

        self.ae = ae = AE()
        ae.acse_timeout = 5
        ae.dimse_timeout = 5
        ae.network_timeout = 5
        ae.add_supported_context(ModalityPerformedProcedureStep)
        scp = ae.start_server(
            ("localhost", 11112), block=False, evt_handlers=[(evt.EVT_N_SET, handle)]
        )

        ae.add_requested_context(ModalityPerformedProcedureStep)
        assoc = ae.associate("localhost", 11112)
        assert assoc.is_established

        ds = Dataset()
        ds.PatientName = "Test^test"
        status, ds = assoc.send_n_set(
            ds, ModalityPerformedProcedureStep, "1.2.840.10008.5.1.1.40.1"
        )
        assert status.Status == 0xFFF0
        assert ds is None
        assoc.release()
        assert assoc.is_released

        scp.shutdown()

    def test_rsp_bad_dataset(self):
        """Test handler returns bad dataset"""

        def handle(event):
            def test():
                pass

            return 0x0000, test

        self.ae = ae = AE()
        ae.add_requested_context(ModalityPerformedProcedureStep)
        ae.add_supported_context(ModalityPerformedProcedureStep)

        handlers = [(evt.EVT_N_SET, handle)]
        scp = ae.start_server(("localhost", 11112), evt_handlers=handlers, block=False)

        ae.acse_timeout = 5
        ae.dimse_timeout = 5
        assoc = ae.associate("localhost", 11112)

        assert assoc.is_established

        # Event Information
        ds = Dataset()
        ds.PatientName = "Test^test"
        status, ds = assoc.send_n_set(
            ds, ModalityPerformedProcedureStep, "1.2.840.10008.5.1.1.40.1"
        )

        assert status.Status == 0x0110
        assert ds is None

        assoc.release()
        scp.shutdown()

    def test_decode_failure(self):
        """Test bad dataset received from peer"""

        def handle(event):
            ds = Dataset()
            ds.PatientName = "Test^test"
            return 0x0000, ds

        self.ae = ae = AE()
        ae.acse_timeout = 5
        ae.dimse_timeout = 0.4
        ae.network_timeout = 5
        ae.add_supported_context(ModalityPerformedProcedureStep)
        scp = ae.start_server(
            ("localhost", 11112), block=False, evt_handlers=[(evt.EVT_N_SET, handle)]
        )

        ae.add_requested_context(ModalityPerformedProcedureStep, ExplicitVRLittleEndian)
        assoc = ae.associate("localhost", 11112)
        assert assoc.is_established

        class DummyReply:
            def getvalue(self):
                def test():
                    pass

                return test

        class DummyMessage:
            is_valid_response = True
            AttributeList = DummyReply()
            Status = 0x0000
            STATUS_OPTIONAL_KEYWORDS = []

        class DummyDIMSE:
            msg_queue = queue.Queue()
            gotten = False

            def send_msg(*args, **kwargs):
                return

            def get_msg(self, *args, **kwargs):
                if not self.gotten:
                    self.gotten = True
                    return 1, DummyMessage()
                return None, None

        assoc._reactor_checkpoint.clear()
        while not assoc._is_paused:
            time.sleep(0.01)
        assoc.dimse = DummyDIMSE()
        assert assoc.is_established
        mod_list = Dataset()
        mod_list.PatientName = "Test^test"
        status, ds = assoc.send_n_set(
            mod_list, ModalityPerformedProcedureStep, "1.2.840.10008.5.1.1.40.1"
        )

        assert status.Status == 0x0110
        assert ds is None

        scp.shutdown()

    def test_extra_status(self):
        """Test extra status elements are available."""

        def handle(event):
            status = Dataset()
            status.Status = 0xFFF0
            status.ErrorComment = "Some comment"
            status.ErrorID = 12
            status.AttributeIdentifierList = [0x00100010]
            return status, event.modification_list

        self.ae = ae = AE()
        ae.acse_timeout = 5
        ae.dimse_timeout = 5
        ae.network_timeout = 5
        ae.add_supported_context(ModalityPerformedProcedureStep)
        scp = ae.start_server(
            ("localhost", 11112), block=False, evt_handlers=[(evt.EVT_N_SET, handle)]
        )

        ae.add_requested_context(ModalityPerformedProcedureStep)
        assoc = ae.associate("localhost", 11112)
        assert assoc.is_established

        ds = Dataset()
        ds.PatientName = "Test^test"
        status, ds = assoc.send_n_set(
            ds, ModalityPerformedProcedureStep, "1.2.840.10008.5.1.1.40.1"
        )
        assert status.Status == 0xFFF0
        assert status.ErrorComment == "Some comment"
        assert status.ErrorID == 12
        assert status.AttributeIdentifierList == Tag(0x00100010)
        assert ds is None
        assoc.release()
        assert assoc.is_released

        scp.shutdown()

    def test_meta_uid(self):
        """Test using a Meta SOP Class"""
        self.ae = ae = AE()
        ae.acse_timeout = 5
        ae.dimse_timeout = 5
        ae.network_timeout = 5
        ae.add_supported_context(BasicGrayscalePrintManagementMeta)
        ae.add_supported_context(Printer)
        scp = ae.start_server(("localhost", 11112), block=False)

        ae.add_requested_context(BasicGrayscalePrintManagementMeta)
        ae.add_requested_context(Printer)
        assoc = ae.associate("localhost", 11112)
        assert assoc.is_established

        assoc.dimse = DummyDIMSE()

        ds = Dataset()
        ds.PatientName = "Test^test"
        # Receives None, None from DummyDIMSE, aborts
        status, ds = assoc.send_n_set(
            ds,
            Printer,
            "1.2.840.10008.5.1.1.40.1",
            meta_uid=BasicGrayscalePrintManagementMeta,
        )
        assert assoc.is_aborted

        scp.shutdown()

        assert assoc.dimse.req.RequestedSOPClassUID == Printer
        assert assoc.dimse.context_id == 1
        assert (
            assoc._accepted_cx[1].abstract_syntax == BasicGrayscalePrintManagementMeta
        )

    def test_meta_uid_good(self):
        """Test sending a request using a Meta SOP Class."""
        handler_data = []

        def handle(event):
            handler_data.append(event)
            return 0x0000, event.modification_list

        self.ae = ae = AE()
        ae.acse_timeout = 5
        ae.dimse_timeout = 5
        ae.network_timeout = 5
        ae.add_supported_context(BasicGrayscalePrintManagementMeta)
        scp = ae.start_server(
            ("localhost", 11112), block=False, evt_handlers=[(evt.EVT_N_SET, handle)]
        )

        ae.add_requested_context(BasicGrayscalePrintManagementMeta)
        assoc = ae.associate("localhost", 11112)
        assert assoc.is_established

        ds = Dataset()
        ds.PatientName = "Test^test"
        status, ds = assoc.send_n_set(
            ds,
            ModalityPerformedProcedureStep,
            "1.2.840.10008.5.1.1.40.1",
            meta_uid=BasicGrayscalePrintManagementMeta,
        )

        assert status.Status == 0x0000
        assert ds is not None
        assert isinstance(ds, Dataset)
        assert status.Status == 0x0000
        assert ds.PatientName == "Test^test"
        assoc.release()
        assert assoc.is_released

        scp.shutdown()

        req = handler_data[0].request
        cx = handler_data[0].context

        assert req.RequestedSOPClassUID == ModalityPerformedProcedureStep
        assert cx.abstract_syntax == BasicGrayscalePrintManagementMeta

    def test_meta_uid_bad(self):
        """Test sending a request using a Meta SOP Class."""

        def handle(event):
            ds = Dataset()
            ds.PatientName = "Test"
            ds.SOPClassUID = DisplaySystem
            ds.SOPInstanceUID = "1.2.3.4"
            return 0x0000, ds

        self.ae = ae = AE()
        ae.acse_timeout = 5
        ae.dimse_timeout = 5
        ae.network_timeout = 5
        ae.add_supported_context(BasicGrayscalePrintManagementMeta)
        scp = ae.start_server(
            ("localhost", 11112), block=False, evt_handlers=[(evt.EVT_N_SET, handle)]
        )

        ae.add_requested_context(BasicGrayscalePrintManagementMeta)
        assoc = ae.associate("localhost", 11112)
        assert assoc.is_established

        ds = Dataset()
        ds.PatientName = "Test^test"
        msg = (
            r"No presentation context for 'Basic Color Print Management "
            r"Meta SOP Class' has been "
            r"accepted by the peer "
            r"for the SCU role"
        )
        with pytest.raises(ValueError, match=msg):
            assoc.send_n_set(
                ds,
                ModalityPerformedProcedureStep,
                "1.2.840.10008.5.1.1.40.1",
                meta_uid=BasicColorPrintManagementMeta,
            )

        assoc.release()
        assert assoc.is_released

        scp.shutdown()


class TestAssociationSendNAction:
    """Run tests on Association send_n_action."""

    def setup_method(self):
        self.ae = None

    def teardown_method(self):
        """Clear any active threads"""
        if self.ae:
            self.ae.shutdown()

    def test_must_be_associated(self):
        """Test can't send without association."""

        def handle(event):
            return 0x0000, event.action_information

        self.ae = ae = AE()
        ae.acse_timeout = 5
        ae.dimse_timeout = 5
        ae.network_timeout = 5
        ae.add_supported_context(ProceduralEventLogging)
        scp = ae.start_server(
            ("localhost", 11112), block=False, evt_handlers=[(evt.EVT_N_ACTION, handle)]
        )

        ae.add_requested_context(ProceduralEventLogging)
        assoc = ae.associate("localhost", 11112)
        assert assoc.is_established

        assoc.release()
        assert assoc.is_released
        assert not assoc.is_established
        with pytest.raises(RuntimeError):
            assoc.send_n_action(None, None, None, None)

        scp.shutdown()

    def test_no_abstract_syntax_match(self):
        """Test SCU when no accepted abstract syntax"""

        def handle(event):
            return 0x0000, event.action_information

        self.ae = ae = AE()
        ae.acse_timeout = 5
        ae.dimse_timeout = 5
        ae.network_timeout = 5
        ae.add_supported_context(ProceduralEventLogging)
        scp = ae.start_server(
            ("localhost", 11112), block=False, evt_handlers=[(evt.EVT_N_ACTION, handle)]
        )

        ae.add_requested_context(ProceduralEventLogging)
        assoc = ae.associate("localhost", 11112)
        assert assoc.is_established

        msg = (
            r"No presentation context for 'Verification SOP Class' has been "
            r"accepted by the peer for the SCU role"
        )
        with pytest.raises(ValueError, match=msg):
            assoc.send_n_action(None, 1, Verification, None)
        assoc.release()
        assert assoc.is_released

        scp.shutdown()

    def test_rq_bad_dataset_raises(self):
        """Test sending bad dataset raises exception."""

        def handle(event):
            return 0x0000, event.action_information

        self.ae = ae = AE()
        ae.acse_timeout = 5
        ae.dimse_timeout = 5
        ae.network_timeout = 5
        ae.add_supported_context(ProceduralEventLogging)
        scp = ae.start_server(
            ("localhost", 11112), block=False, evt_handlers=[(evt.EVT_N_ACTION, handle)]
        )

        ae.add_requested_context(ProceduralEventLogging, ExplicitVRLittleEndian)
        assoc = ae.associate("localhost", 11112)
        assert assoc.is_established

        ds = Dataset()
        ds.PerimeterValue = b"\x00\x01"
        msg = r"Failed to encode the supplied 'Action Information' dataset"
        with pytest.raises(ValueError, match=msg):
            assoc.send_n_action(ds, 1, ProceduralEventLogging, "1.2.3")
        assoc.release()
        assert assoc.is_released

        scp.shutdown()

    def test_rsp_none(self):
        """Test no response from peer"""

        def handle(event):
            time.sleep(5)
            return 0x0000, event.action_information

        self.ae = ae = AE()
        ae.acse_timeout = 5
        ae.dimse_timeout = 0.2
        ae.network_timeout = 5
        ae.add_supported_context(ProceduralEventLogging)
        scp = ae.start_server(
            ("localhost", 11112), block=False, evt_handlers=[(evt.EVT_N_ACTION, handle)]
        )

        ae.add_requested_context(ProceduralEventLogging)
        assoc = ae.associate("localhost", 11112)
        assert assoc.is_established

        class DummyDIMSE:
            msg_queue = queue.Queue()

            def send_msg(*args, **kwargs):
                return

            def get_msg(*args, **kwargs):
                return None, None

        assoc._reactor_checkpoint.clear()
        while not assoc._is_paused:
            time.sleep(0.01)
        assoc.dimse = DummyDIMSE()
        ds = Dataset()
        ds.PatientName = "Test^test"
        status, ds = assoc.send_n_action(
            ds, 1, ProceduralEventLogging, "1.2.840.10008.5.1.1.40.1"
        )

        assert status == Dataset()
        assert ds is None
        assert assoc.is_aborted

        scp.shutdown()

    def test_rsp_invalid(self):
        """Test invalid DIMSE message received from peer"""

        def handle(event):
            return 0x0000, event.action_information

        self.ae = ae = AE()
        ae.acse_timeout = 5
        ae.dimse_timeout = 5
        ae.network_timeout = 5
        ae.add_supported_context(ProceduralEventLogging)
        scp = ae.start_server(
            ("localhost", 11112), block=False, evt_handlers=[(evt.EVT_N_ACTION, handle)]
        )

        ae.add_requested_context(ProceduralEventLogging)
        assoc = ae.associate("localhost", 11112)
        assert assoc.is_established

        class DummyResponse:
            is_valid_response = False

        class DummyDIMSE:
            msg_queue = queue.Queue()
            gotten = False

            def send_msg(*args, **kwargs):
                return

            def get_msg(self, *args, **kwargs):
                if not self.gotten:
                    self.gotten = True
                    return None, DummyResponse()
                return None, None

        assoc._reactor_checkpoint.clear()
        while not assoc._is_paused:
            time.sleep(0.01)
        assoc.dimse = DummyDIMSE()
        ds = Dataset()
        ds.PatientName = "Test^test"
        status, ds = assoc.send_n_action(
            ds, 1, ProceduralEventLogging, "1.2.840.10008.5.1.1.40.1"
        )
        assert status == Dataset()
        assert ds is None
        assert assoc.is_aborted

        scp.shutdown()

    def test_rsp_failure(self):
        """Test receiving a failure response from the peer"""

        def handle(event):
            return 0x0112, None

        self.ae = ae = AE()
        ae.acse_timeout = 5
        ae.dimse_timeout = 5
        ae.network_timeout = 5
        ae.add_supported_context(ProceduralEventLogging)
        scp = ae.start_server(
            ("localhost", 11112), block=False, evt_handlers=[(evt.EVT_N_ACTION, handle)]
        )

        ae.add_requested_context(ProceduralEventLogging)
        assoc = ae.associate("localhost", 11112)
        assert assoc.is_established

        ds = Dataset()
        ds.PatientName = "Test^test"
        status, ds = assoc.send_n_action(
            ds, 1, ProceduralEventLogging, "1.2.840.10008.5.1.1.40.1"
        )
        assert status.Status == 0x0112
        assert ds is None
        assoc.release()
        assert assoc.is_released

        scp.shutdown()

    def test_rsp_warning(self):
        """Test receiving a warning response from the peer"""

        def handle(event):
            return 0x0116, event.action_information

        self.ae = ae = AE()
        ae.acse_timeout = 5
        ae.dimse_timeout = 5
        ae.network_timeout = 5
        ae.add_supported_context(ProceduralEventLogging)
        scp = ae.start_server(
            ("localhost", 11112), block=False, evt_handlers=[(evt.EVT_N_ACTION, handle)]
        )

        ae.add_requested_context(ProceduralEventLogging)
        assoc = ae.associate("localhost", 11112)
        assert assoc.is_established

        ds = Dataset()
        ds.PatientName = "Test^test"
        status, ds = assoc.send_n_action(
            ds, 1, ProceduralEventLogging, "1.2.840.10008.5.1.1.40.1"
        )
        assert status.Status == 0x0116
        assert ds.PatientName == "Test^test"
        assoc.release()
        assert assoc.is_released

        scp.shutdown()

    def test_rsp_success(self):
        """Test receiving a success response from the peer"""

        def handle(event):
            return 0x0000, event.action_information

        self.ae = ae = AE()
        ae.acse_timeout = 5
        ae.dimse_timeout = 5
        ae.network_timeout = 5
        ae.add_supported_context(ProceduralEventLogging)
        scp = ae.start_server(
            ("localhost", 11112), block=False, evt_handlers=[(evt.EVT_N_ACTION, handle)]
        )

        ae.add_requested_context(ProceduralEventLogging)
        assoc = ae.associate("localhost", 11112)
        assert assoc.is_established

        ds = Dataset()
        ds.PatientName = "Test^test"
        status, ds = assoc.send_n_action(
            ds, 1, ProceduralEventLogging, "1.2.840.10008.5.1.1.40.1"
        )
        assert status.Status == 0x0000
        assert ds.PatientName == "Test^test"
        assoc.release()
        assert assoc.is_released

        scp.shutdown()

    def test_rsp_unknown_status(self):
        """Test unknown status value returned by peer"""

        def handle(event):
            return 0xFFF0, event.action_information

        self.ae = ae = AE()
        ae.acse_timeout = 5
        ae.dimse_timeout = 5
        ae.network_timeout = 5
        ae.add_supported_context(ProceduralEventLogging)
        scp = ae.start_server(
            ("localhost", 11112), block=False, evt_handlers=[(evt.EVT_N_ACTION, handle)]
        )

        ae.add_requested_context(ProceduralEventLogging)
        assoc = ae.associate("localhost", 11112)
        assert assoc.is_established

        ds = Dataset()
        ds.PatientName = "Test^test"
        status, ds = assoc.send_n_action(
            ds, 1, ProceduralEventLogging, "1.2.840.10008.5.1.1.40.1"
        )
        assert status.Status == 0xFFF0
        assert ds is None
        assoc.release()
        assert assoc.is_released

        scp.shutdown()

    def test_rsp_bad_dataset(self):
        """Test bad dataset received from peer"""

        def handle(event):
            return 0x0000, event.action_information

        self.ae = ae = AE()
        ae.acse_timeout = 5
        ae.dimse_timeout = 5
        ae.network_timeout = 5
        ae.add_supported_context(PrintJob)
        scp = ae.start_server(
            ("localhost", 11112), block=False, evt_handlers=[(evt.EVT_N_ACTION, handle)]
        )

        ae.add_requested_context(PrintJob)
        assoc = ae.associate("localhost", 11112)
        assert assoc.is_established

        class DummyReply:
            def getvalue(self):
                def test():
                    pass

                return test

        class DummyMessage:
            is_valid_response = True
            is_valid_request = False
            msg_type = None
            ActionReply = DummyReply()
            Status = 0x0000
            STATUS_OPTIONAL_KEYWORDS = []

        class DummyDIMSE:
            msg_queue = queue.Queue()

            def send_msg(*args, **kwargs):
                return

            def get_msg(*args, **kwargs):
                rsp = DummyMessage()
                return 1, rsp

        assoc._reactor_checkpoint.clear()
        while not assoc._is_paused:
            time.sleep(0.01)
        assoc.dimse = DummyDIMSE()
        ds = Dataset()
        ds.PatientName = "Test^test"
        status, ds = assoc.send_n_action(ds, 1, PrintJob, "1.2.840.10008.5.1.1.40.1")

        assert status.Status == 0x0110
        assert ds is None

        scp.shutdown()

    def test_extra_status(self):
        """Test extra status elements are available."""

        def handle(event):
            ds = Dataset()
            ds.Status = 0xFFF0
            ds.ErrorComment = "Some comment"
            ds.ErrorID = 12
            return ds, event.action_information

        self.ae = ae = AE()
        ae.acse_timeout = 5
        ae.dimse_timeout = 5
        ae.network_timeout = 5
        ae.add_supported_context(PrintJob)
        scp = ae.start_server(
            ("localhost", 11112), block=False, evt_handlers=[(evt.EVT_N_ACTION, handle)]
        )

        ae.add_requested_context(PrintJob)
        assoc = ae.associate("localhost", 11112)
        assert assoc.is_established

        ds = Dataset()
        ds.PatientName = "Test^test"
        status, ds = assoc.send_n_action(ds, 1, PrintJob, "1.2.840.10008.5.1.1.40.1")
        assert status.Status == 0xFFF0
        assert status.ErrorComment == "Some comment"
        assert status.ErrorID == 12
        assert ds is None
        assoc.release()
        assert assoc.is_released

        scp.shutdown()

    def test_meta_uid(self):
        """Test using a Meta SOP Class"""
        self.ae = ae = AE()
        ae.acse_timeout = 5
        ae.dimse_timeout = 5
        ae.network_timeout = 5
        ae.add_supported_context(BasicGrayscalePrintManagementMeta)
        ae.add_supported_context(Printer)
        scp = ae.start_server(("localhost", 11112), block=False)

        ae.add_requested_context(BasicGrayscalePrintManagementMeta)
        ae.add_requested_context(Printer)
        assoc = ae.associate("localhost", 11112)
        assert assoc.is_established

        assoc.dimse = DummyDIMSE()

        ds = Dataset()
        ds.PatientName = "Test^test"
        # Receives None, None from DummyDIMSE, aborts
        status, ds = assoc.send_n_action(
            ds,
            1,
            Printer,
            "1.2.840.10008.5.1.1.40.1",
            meta_uid=BasicGrayscalePrintManagementMeta,
        )
        assert assoc.is_aborted

        scp.shutdown()

        assert assoc.dimse.req.RequestedSOPClassUID == Printer
        assert assoc.dimse.context_id == 1
        assert (
            assoc._accepted_cx[1].abstract_syntax == BasicGrayscalePrintManagementMeta
        )

    def test_meta_uid_good(self):
        """Test sending a request using a Meta SOP Class."""
        handler_data = []

        def handle(event):
            handler_data.append(event)
            return 0x0000, event.action_information

        self.ae = ae = AE()
        ae.acse_timeout = 5
        ae.dimse_timeout = 5
        ae.network_timeout = 5
        ae.add_supported_context(BasicGrayscalePrintManagementMeta)
        scp = ae.start_server(
            ("localhost", 11112), block=False, evt_handlers=[(evt.EVT_N_ACTION, handle)]
        )

        ae.add_requested_context(BasicGrayscalePrintManagementMeta)
        assoc = ae.associate("localhost", 11112)
        assert assoc.is_established

        ds = Dataset()
        ds.PatientName = "Test^test"
        status, ds = assoc.send_n_action(
            ds,
            1,
            ProceduralEventLogging,
            "1.2.840.10008.5.1.1.40.1",
            meta_uid=BasicGrayscalePrintManagementMeta,
        )
        assert status.Status == 0x0000
        assert ds.PatientName == "Test^test"
        assoc.release()
        assert assoc.is_released

        scp.shutdown()

        req = handler_data[0].request
        cx = handler_data[0].context

        assert req.RequestedSOPClassUID == ProceduralEventLogging
        assert cx.abstract_syntax == BasicGrayscalePrintManagementMeta

    def test_meta_uid_bad(self):
        """Test sending a request using a Meta SOP Class."""

        def handle(event):
            ds = Dataset()
            ds.PatientName = "Test"
            ds.SOPClassUID = DisplaySystem
            ds.SOPInstanceUID = "1.2.3.4"
            return 0x0000, ds

        self.ae = ae = AE()
        ae.acse_timeout = 5
        ae.dimse_timeout = 5
        ae.network_timeout = 5
        ae.add_supported_context(BasicGrayscalePrintManagementMeta)
        scp = ae.start_server(
            ("localhost", 11112), block=False, evt_handlers=[(evt.EVT_N_ACTION, handle)]
        )

        ae.add_requested_context(BasicGrayscalePrintManagementMeta)
        assoc = ae.associate("localhost", 11112)
        assert assoc.is_established

        ds = Dataset()
        ds.PatientName = "Test^test"
        msg = (
            r"No presentation context for 'Basic Color Print Management "
            r"Meta SOP Class' has been "
            r"accepted by the peer "
            r"for the SCU role"
        )
        with pytest.raises(ValueError, match=msg):
            assoc.send_n_action(
                ds,
                1,
                ProceduralEventLogging,
                "1.2.840.10008.5.1.1.40.1",
                meta_uid=BasicColorPrintManagementMeta,
            )

        assoc.release()
        assert assoc.is_released

        scp.shutdown()


class TestAssociationSendNCreate:
    """Run tests on Association send_n_create."""

    def setup_method(self):
        self.ae = None

    def teardown_method(self):
        """Clear any active threads"""
        if self.ae:
            self.ae.shutdown()

    def test_must_be_associated(self):
        """Test can't send without association."""

        def handle(event):
            return 0x0000, Dataset()

        self.ae = ae = AE()
        ae.acse_timeout = 5
        ae.dimse_timeout = 5
        ae.network_timeout = 5
        ae.add_supported_context(ModalityPerformedProcedureStep)

        handlers = [(evt.EVT_N_CREATE, handle)]
        scp = ae.start_server(("localhost", 11112), evt_handlers=handlers, block=False)

        ae.add_requested_context(ModalityPerformedProcedureStep)
        assoc = ae.associate("localhost", 11112)
        assert assoc.is_established

        assoc.release()
        assert assoc.is_released
        assert not assoc.is_established
        with pytest.raises(RuntimeError):
            assoc.send_n_create(None, None, None)

        scp.shutdown()

    def test_no_abstract_syntax_match(self):
        """Test SCU when no accepted abstract syntax"""

        def handle(event):
            return 0x0000, Dataset()

        self.ae = ae = AE()
        ae.acse_timeout = 5
        ae.dimse_timeout = 5
        ae.network_timeout = 5
        ae.add_supported_context(ModalityPerformedProcedureStep)

        handlers = [(evt.EVT_N_CREATE, handle)]
        scp = ae.start_server(("localhost", 11112), evt_handlers=handlers, block=False)

        ae.add_requested_context(ModalityPerformedProcedureStep)
        assoc = ae.associate("localhost", 11112)
        assert assoc.is_established

        msg = (
            r"No presentation context for 'Verification SOP Class' has been "
            r"accepted by the peer for the SCU role"
        )
        with pytest.raises(ValueError, match=msg):
            assoc.send_n_create(None, Verification, None)
        assoc.release()
        assert assoc.is_released

        scp.shutdown()

    def test_rq_bad_dataset_raises(self):
        """Test sending bad dataset raises exception."""

        def handle(event):
            return 0x0000, Dataset()

        self.ae = ae = AE()
        ae.acse_timeout = 5
        ae.dimse_timeout = 5
        ae.network_timeout = 5
        ae.add_supported_context(ModalityPerformedProcedureStep)

        handlers = [(evt.EVT_N_CREATE, handle)]
        scp = ae.start_server(("localhost", 11112), evt_handlers=handlers, block=False)

        ae.add_requested_context(ModalityPerformedProcedureStep, ExplicitVRLittleEndian)
        assoc = ae.associate("localhost", 11112)
        assert assoc.is_established

        ds = Dataset()
        ds.PerimeterValue = b"\x00\x01"
        msg = r"Failed to encode the supplied 'Attribute List' dataset"
        with pytest.raises(ValueError, match=msg):
            assoc.send_n_create(ds, ModalityPerformedProcedureStep, "1.2.3")
        assoc.release()
        assert assoc.is_released

        scp.shutdown()

    def test_rsp_none(self):
        """Test no response from peer"""

        def handle(event):
            time.sleep(5)
            return 0x0000, Dataset()

        self.ae = ae = AE()
        ae.acse_timeout = 5
        ae.dimse_timeout = 0.2
        ae.network_timeout = 5
        ae.add_supported_context(ModalityPerformedProcedureStep)

        handlers = [(evt.EVT_N_CREATE, handle)]
        scp = ae.start_server(("localhost", 11112), evt_handlers=handlers, block=False)

        ae.add_requested_context(ModalityPerformedProcedureStep)
        assoc = ae.associate("localhost", 11112)
        assert assoc.is_established

        ds = Dataset()
        ds.PatientName = "Test^test"
        status, ds = assoc.send_n_create(
            ds, ModalityPerformedProcedureStep, "1.2.840.10008.5.1.1.40.1"
        )

        assert status == Dataset()
        assert ds is None
        assert assoc.is_aborted

        scp.shutdown()

    def test_rsp_invalid(self):
        """Test invalid DIMSE message received from peer"""

        def handle(event):
            return 0x0000, Dataset()

        self.ae = ae = AE()
        ae.acse_timeout = 5
        ae.dimse_timeout = 5
        ae.network_timeout = 5
        ae.add_supported_context(ModalityPerformedProcedureStep)

        handlers = [(evt.EVT_N_CREATE, handle)]
        scp = ae.start_server(("localhost", 11112), evt_handlers=handlers, block=False)

        ae.add_requested_context(ModalityPerformedProcedureStep)
        assoc = ae.associate("localhost", 11112)
        assert assoc.is_established

        class DummyResponse:
            is_valid_response = False

        class DummyDIMSE:
            msg_queue = queue.Queue()
            gotten = False

            def send_msg(*args, **kwargs):
                return

            def get_msg(self, *args, **kwargs):
                if not self.gotten:
                    self.gotten = True
                    return None, DummyResponse()
                return None, None

        assoc._reactor_checkpoint.clear()
        while not assoc._is_paused:
            time.sleep(0.01)
        assoc.dimse = DummyDIMSE()
        ds = Dataset()
        ds.PatientName = "Test^test"
        status, ds = assoc.send_n_create(
            ds, ModalityPerformedProcedureStep, "1.2.840.10008.5.1.1.40.1"
        )
        assert status == Dataset()
        assert ds is None
        assert assoc.is_aborted

        scp.shutdown()

    def test_rsp_failure(self):
        """Test receiving a failure response from the peer"""

        def handle(event):
            return 0x0112, Dataset()

        self.ae = ae = AE()
        ae.acse_timeout = 5
        ae.dimse_timeout = 5
        ae.network_timeout = 5
        ae.add_supported_context(ModalityPerformedProcedureStep)

        handlers = [(evt.EVT_N_CREATE, handle)]
        scp = ae.start_server(("localhost", 11112), evt_handlers=handlers, block=False)

        ae.add_requested_context(ModalityPerformedProcedureStep)
        assoc = ae.associate("localhost", 11112)
        assert assoc.is_established

        ds = Dataset()
        ds.PatientName = "Test^test"
        status, ds = assoc.send_n_create(
            ds, ModalityPerformedProcedureStep, "1.2.840.10008.5.1.1.40.1"
        )
        assert status.Status == 0x0112
        assert ds is None
        assoc.release()
        assert assoc.is_released

        scp.shutdown()

    def test_rsp_warning(self):
        """Test receiving a warning response from the peer"""

        def handle(event):
            return 0x0116, event.attribute_list

        self.ae = ae = AE()
        ae.acse_timeout = 5
        ae.dimse_timeout = 5
        ae.network_timeout = 5
        ae.add_supported_context(ModalityPerformedProcedureStep)

        handlers = [(evt.EVT_N_CREATE, handle)]
        scp = ae.start_server(("localhost", 11112), evt_handlers=handlers, block=False)

        ae.add_requested_context(ModalityPerformedProcedureStep)
        assoc = ae.associate("localhost", 11112)
        assert assoc.is_established

        ds = Dataset()
        ds.PatientName = "Test^test"
        status, ds = assoc.send_n_create(
            ds, ModalityPerformedProcedureStep, "1.2.840.10008.5.1.1.40.1"
        )
        assert status.Status == 0x0116
        assert ds.PatientName == "Test^test"
        assoc.release()
        assert assoc.is_released

        scp.shutdown()

    def test_rsp_success(self):
        """Test receiving a success response from the peer"""

        def handle(event):
            return 0x0000, event.attribute_list

        self.ae = ae = AE()
        ae.acse_timeout = 5
        ae.dimse_timeout = 5
        ae.network_timeout = 5
        ae.add_supported_context(ModalityPerformedProcedureStep)

        handlers = [(evt.EVT_N_CREATE, handle)]
        scp = ae.start_server(("localhost", 11112), evt_handlers=handlers, block=False)

        ae.add_requested_context(ModalityPerformedProcedureStep)
        assoc = ae.associate("localhost", 11112)
        assert assoc.is_established

        ds = Dataset()
        ds.PatientName = "Test^test"
        status, ds = assoc.send_n_create(
            ds, ModalityPerformedProcedureStep, "1.2.840.10008.5.1.1.40.1"
        )
        assert status.Status == 0x0000
        assert ds.PatientName == "Test^test"
        assoc.release()
        assert assoc.is_released

        scp.shutdown()

    def test_rsp_unknown_status(self):
        """Test unknown status value returned by peer"""

        def handle(event):
            return 0xFFF0, event.attribute_list

        self.ae = ae = AE()
        ae.acse_timeout = 5
        ae.dimse_timeout = 5
        ae.network_timeout = 5
        ae.add_supported_context(ModalityPerformedProcedureStep)

        handlers = [(evt.EVT_N_CREATE, handle)]
        scp = ae.start_server(("localhost", 11112), evt_handlers=handlers, block=False)

        ae.add_requested_context(ModalityPerformedProcedureStep)
        assoc = ae.associate("localhost", 11112)
        assert assoc.is_established

        ds = Dataset()
        ds.PatientName = "Test^test"
        status, ds = assoc.send_n_create(
            ds, ModalityPerformedProcedureStep, "1.2.840.10008.5.1.1.40.1"
        )
        assert status.Status == 0xFFF0
        assert ds is None
        assoc.release()
        assert assoc.is_released

        scp.shutdown()

    def test_rsp_bad_dataset(self):
        """Test handler returns bad dataset"""

        def handle(event):
            def test():
                pass

            return 0x0000, test

        self.ae = ae = AE()
        ae.add_requested_context(ModalityPerformedProcedureStep)
        ae.add_supported_context(ModalityPerformedProcedureStep)

        handlers = [(evt.EVT_N_CREATE, handle)]
        scp = ae.start_server(("localhost", 11112), evt_handlers=handlers, block=False)

        ae.acse_timeout = 5
        ae.dimse_timeout = 5
        assoc = ae.associate("localhost", 11112)

        assert assoc.is_established

        # Event Information
        ds = Dataset()
        ds.PatientName = "Test^test"
        status, ds = assoc.send_n_create(
            ds, ModalityPerformedProcedureStep, "1.2.840.10008.5.1.1.40.1"
        )

        assert status.Status == 0x0110
        assert ds is None

        assoc.release()
        scp.shutdown()

    def test_decode_failure(self):
        """Test bad dataset received from peer"""

        def handle(event):
            return 0x0000, event.attribute_list

        self.ae = ae = AE()
        ae.acse_timeout = 5
        ae.dimse_timeout = 5
        ae.network_timeout = 5
        ae.add_supported_context(ModalityPerformedProcedureStep)

        handlers = [(evt.EVT_N_CREATE, handle)]
        scp = ae.start_server(("localhost", 11112), evt_handlers=handlers, block=False)

        ae.add_requested_context(ModalityPerformedProcedureStep)
        assoc = ae.associate("localhost", 11112)
        assert assoc.is_established

        class DummyReply:
            def getvalue(self):
                def test():
                    pass

                return test

        class DummyMessage:
            is_valid_response = True
            is_valid_request = False
            AttributeList = DummyReply()
            Status = 0x0000
            STATUS_OPTIONAL_KEYWORDS = []

        class DummyDIMSE:
            msg_queue = queue.Queue()
            gotten = False

            def send_msg(*args, **kwargs):
                return

            def get_msg(self, *args, **kwargs):
                if not self.gotten:
                    self.gotten = True
                    return 1, DummyMessage()
                return None, None

        assoc._reactor_checkpoint.clear()
        while not assoc._is_paused:
            time.sleep(0.01)
        assoc.dimse = DummyDIMSE()
        ds = Dataset()
        ds.PatientName = "Test^test"
        status, ds = assoc.send_n_create(
            ds, ModalityPerformedProcedureStep, "1.2.840.10008.5.1.1.40.1"
        )
        assert status.Status == 0x0110
        assert ds is None
        assoc.release()
        assert assoc.is_released

        scp.shutdown()

    def test_extra_status(self):
        """Test extra status elements are available."""

        def handle(event):
            status = Dataset()
            status.Status = 0xFFF0
            status.ErrorComment = "Some comment"
            status.ErrorID = 12
            return status, event.attribute_list

        self.ae = ae = AE()
        ae.acse_timeout = 5
        ae.dimse_timeout = 5
        ae.network_timeout = 5
        ae.add_supported_context(ModalityPerformedProcedureStep)

        handlers = [(evt.EVT_N_CREATE, handle)]
        scp = ae.start_server(("localhost", 11112), evt_handlers=handlers, block=False)

        ae.add_requested_context(ModalityPerformedProcedureStep)
        assoc = ae.associate("localhost", 11112)
        assert assoc.is_established

        ds = Dataset()
        ds.PatientName = "Test^test"
        status, ds = assoc.send_n_create(
            ds, ModalityPerformedProcedureStep, "1.2.840.10008.5.1.1.40.1"
        )
        assert status.Status == 0xFFF0
        assert status.ErrorComment == "Some comment"
        assert status.ErrorID == 12
        assert ds is None
        assoc.release()
        assert assoc.is_released

        scp.shutdown()

    def test_meta_uid(self):
        """Test using a Meta SOP Class"""
        self.ae = ae = AE()
        ae.acse_timeout = 5
        ae.dimse_timeout = 5
        ae.network_timeout = 5
        ae.add_supported_context(BasicGrayscalePrintManagementMeta)
        ae.add_supported_context(Printer)
        scp = ae.start_server(("localhost", 11112), block=False)

        ae.add_requested_context(BasicGrayscalePrintManagementMeta)
        ae.add_requested_context(Printer)
        assoc = ae.associate("localhost", 11112)
        assert assoc.is_established

        assoc.dimse = DummyDIMSE()

        ds = Dataset()
        ds.PatientName = "Test^test"
        # Receives None, None from DummyDIMSE, aborts
        status, ds = assoc.send_n_create(
            ds,
            Printer,
            "1.2.840.10008.5.1.1.40.1",
            meta_uid=BasicGrayscalePrintManagementMeta,
        )
        assert assoc.is_aborted

        scp.shutdown()

        assert assoc.dimse.req.AffectedSOPClassUID == Printer
        assert assoc.dimse.context_id == 1
        assert (
            assoc._accepted_cx[1].abstract_syntax == BasicGrayscalePrintManagementMeta
        )

    def test_meta_uid_good(self):
        """Test sending a request using a Meta SOP Class."""
        handler_data = []

        def handle(event):
            handler_data.append(event)
            return 0x0000, event.attribute_list

        self.ae = ae = AE()
        ae.acse_timeout = 5
        ae.dimse_timeout = 5
        ae.network_timeout = 5
        ae.add_supported_context(BasicGrayscalePrintManagementMeta)
        scp = ae.start_server(
            ("localhost", 11112), block=False, evt_handlers=[(evt.EVT_N_CREATE, handle)]
        )

        ae.add_requested_context(BasicGrayscalePrintManagementMeta)
        assoc = ae.associate("localhost", 11112)
        assert assoc.is_established

        ds = Dataset()
        ds.PatientName = "Test^test"
        status, ds = assoc.send_n_create(
            ds,
            ModalityPerformedProcedureStep,
            "1.2.840.10008.5.1.1.40.1",
            meta_uid=BasicGrayscalePrintManagementMeta,
        )
        assert status.Status == 0x0000
        assert ds.PatientName == "Test^test"
        assoc.release()
        assert assoc.is_released

        scp.shutdown()

        req = handler_data[0].request
        cx = handler_data[0].context

        assert req.AffectedSOPClassUID == ModalityPerformedProcedureStep
        assert cx.abstract_syntax == BasicGrayscalePrintManagementMeta

    def test_meta_uid_bad(self):
        """Test sending a request using a Meta SOP Class."""

        def handle(event):
            ds = Dataset()
            ds.PatientName = "Test"
            ds.SOPClassUID = DisplaySystem
            ds.SOPInstanceUID = "1.2.3.4"
            return 0x0000, ds

        self.ae = ae = AE()
        ae.acse_timeout = 5
        ae.dimse_timeout = 5
        ae.network_timeout = 5
        ae.add_supported_context(BasicGrayscalePrintManagementMeta)
        scp = ae.start_server(
            ("localhost", 11112), block=False, evt_handlers=[(evt.EVT_N_CREATE, handle)]
        )

        ae.add_requested_context(BasicGrayscalePrintManagementMeta)
        assoc = ae.associate("localhost", 11112)
        assert assoc.is_established

        ds = Dataset()
        ds.PatientName = "Test^test"
        msg = (
            r"No presentation context for 'Basic Color Print Management "
            r"Meta SOP Class' has been "
            r"accepted by the peer "
            r"for the SCU role"
        )
        with pytest.raises(ValueError, match=msg):
            assoc.send_n_create(
                ds,
                ModalityPerformedProcedureStep,
                "1.2.840.10008.5.1.1.40.1",
                meta_uid=BasicColorPrintManagementMeta,
            )

        assoc.release()
        assert assoc.is_released

        scp.shutdown()


class TestAssociationSendNDelete:
    """Run tests on Association send_n_delete."""

    def setup_method(self):
        self.ae = None

    def teardown_method(self):
        """Clear any active threads"""
        if self.ae:
            self.ae.shutdown()

    def test_must_be_associated(self):
        """Test can't send without association."""

        def handle(event):
            return 0x0000

        self.ae = ae = AE()
        ae.acse_timeout = 5
        ae.dimse_timeout = 5
        ae.network_timeout = 5
        ae.add_supported_context(BasicFilmSession)
        scp = ae.start_server(
            ("localhost", 11112), block=False, evt_handlers=[(evt.EVT_N_DELETE, handle)]
        )

        ae.add_requested_context(BasicFilmSession)
        assoc = ae.associate("localhost", 11112)
        assert assoc.is_established

        assoc.release()
        assert assoc.is_released
        assert not assoc.is_established
        with pytest.raises(RuntimeError):
            assoc.send_n_delete(None, None)

        scp.shutdown()

    def test_no_abstract_syntax_match(self):
        """Test SCU when no accepted abstract syntax"""

        def handle(event):
            return 0x0000

        self.ae = ae = AE()
        ae.acse_timeout = 5
        ae.dimse_timeout = 5
        ae.network_timeout = 5
        ae.add_supported_context(BasicFilmSession)
        scp = ae.start_server(
            ("localhost", 11112), block=False, evt_handlers=[(evt.EVT_N_DELETE, handle)]
        )

        ae.add_requested_context(BasicFilmSession)
        assoc = ae.associate("localhost", 11112)
        assert assoc.is_established

        msg = (
            r"No presentation context for 'Verification SOP Class' has been "
            r"accepted by the peer for the SCU role"
        )
        with pytest.raises(ValueError, match=msg):
            assoc.send_n_delete(Verification, None)
        assoc.release()
        assert assoc.is_released

        scp.shutdown()

    def test_rsp_none(self):
        """Test no response from peer"""

        def handle(event):
            time.sleep(5)
            return 0x0000

        self.ae = ae = AE()
        ae.acse_timeout = 5
        ae.dimse_timeout = 0.2
        ae.network_timeout = 5
        ae.add_supported_context(BasicFilmSession)
        scp = ae.start_server(
            ("localhost", 11112), block=False, evt_handlers=[(evt.EVT_N_DELETE, handle)]
        )

        ae.add_requested_context(BasicFilmSession)
        assoc = ae.associate("localhost", 11112)
        assert assoc.is_established

        status = assoc.send_n_delete(BasicFilmSession, "1.2.840.10008.5.1.1.40.1")
        assert status == Dataset()
        assert assoc.is_aborted

        scp.shutdown()

    def test_rsp_invalid(self):
        """Test invalid DIMSE message received from peer"""

        def handle(event):
            return 0x0000

        self.ae = ae = AE()
        ae.acse_timeout = 5
        ae.dimse_timeout = 5
        ae.network_timeout = 5
        ae.add_supported_context(BasicFilmSession)
        scp = ae.start_server(
            ("localhost", 11112), block=False, evt_handlers=[(evt.EVT_N_DELETE, handle)]
        )

        ae.add_requested_context(BasicFilmSession)
        assoc = ae.associate("localhost", 11112)
        assert assoc.is_established

        class DummyResponse:
            is_valid_response = False

        class DummyDIMSE:
            msg_queue = queue.Queue()
            gotten = False

            def send_msg(*args, **kwargs):
                return

            def get_msg(self, *args, **kwargs):
                if not self.gotten:
                    self.gotten = True
                    return None, DummyResponse()
                return None, None

        assoc._reactor_checkpoint.clear()
        while not assoc._is_paused:
            time.sleep(0.01)
        assoc.dimse = DummyDIMSE()
        status = assoc.send_n_delete(BasicFilmSession, "1.2.840.10008.5.1.1.40.1")
        assert status == Dataset()
        assert assoc.is_aborted

        scp.shutdown()

    def test_rsp_failure(self):
        """Test receiving a failure response from the peer"""

        def handle(event):
            return 0x0112

        self.ae = ae = AE()
        ae.acse_timeout = 5
        ae.dimse_timeout = 5
        ae.network_timeout = 5
        ae.add_supported_context(BasicFilmSession)
        scp = ae.start_server(
            ("localhost", 11112), block=False, evt_handlers=[(evt.EVT_N_DELETE, handle)]
        )

        ae.add_requested_context(BasicFilmSession)
        assoc = ae.associate("localhost", 11112)
        assert assoc.is_established

        status = assoc.send_n_delete(BasicFilmSession, "1.2.840.10008.5.1.1.40.1")
        assert status.Status == 0x0112
        assoc.release()
        assert assoc.is_released

        scp.shutdown()

    def test_rsp_success(self):
        """Test receiving a success response from the peer"""

        def handle(event):
            return 0x0000

        self.ae = ae = AE()
        ae.acse_timeout = 5
        ae.dimse_timeout = 5
        ae.network_timeout = 5
        ae.add_supported_context(BasicFilmSession)
        scp = ae.start_server(
            ("localhost", 11112), block=False, evt_handlers=[(evt.EVT_N_DELETE, handle)]
        )

        ae.add_requested_context(BasicFilmSession)
        assoc = ae.associate("localhost", 11112)
        assert assoc.is_established

        status = assoc.send_n_delete(BasicFilmSession, "1.2.840.10008.5.1.1.40.1")
        assert status.Status == 0x0000
        assoc.release()
        assert assoc.is_released

        scp.shutdown()

    def test_rsp_unknown_status(self):
        """Test unknown status value returned by peer"""

        def handle(event):
            return 0xFFF0

        self.ae = ae = AE()
        ae.acse_timeout = 5
        ae.dimse_timeout = 5
        ae.network_timeout = 5
        ae.add_supported_context(BasicFilmSession)
        scp = ae.start_server(
            ("localhost", 11112), block=False, evt_handlers=[(evt.EVT_N_DELETE, handle)]
        )

        ae.add_requested_context(BasicFilmSession)
        assoc = ae.associate("localhost", 11112)
        assert assoc.is_established

        status = assoc.send_n_delete(BasicFilmSession, "1.2.840.10008.5.1.1.40.1")
        assert status.Status == 0xFFF0
        assoc.release()
        assert assoc.is_released

        scp.shutdown()

    def test_extra_status(self):
        """Test extra status elements are available."""

        def handle(event):
            ds = Dataset()
            ds.Status = 0xFFF0
            ds.ErrorComment = "Some comment"
            ds.ErrorID = 12
            return ds

        self.ae = ae = AE()
        ae.acse_timeout = 5
        ae.dimse_timeout = 5
        ae.network_timeout = 5
        ae.add_supported_context(BasicFilmSession)
        scp = ae.start_server(
            ("localhost", 11112), block=False, evt_handlers=[(evt.EVT_N_DELETE, handle)]
        )

        ae.add_requested_context(BasicFilmSession)
        assoc = ae.associate("localhost", 11112)
        assert assoc.is_established

        status = assoc.send_n_delete(BasicFilmSession, "1.2.840.10008.5.1.1.40.1")
        assert status.Status == 0xFFF0
        assert status.ErrorComment == "Some comment"
        assert status.ErrorID == 12
        assoc.release()
        assert assoc.is_released

        scp.shutdown()

    def test_meta_uid(self):
        """Test using a Meta SOP Class"""
        self.ae = ae = AE()
        ae.acse_timeout = 5
        ae.dimse_timeout = 5
        ae.network_timeout = 5
        ae.add_supported_context(BasicGrayscalePrintManagementMeta)
        ae.add_supported_context(Printer)
        scp = ae.start_server(("localhost", 11112), block=False)

        ae.add_requested_context(BasicGrayscalePrintManagementMeta)
        ae.add_requested_context(Printer)
        assoc = ae.associate("localhost", 11112)
        assert assoc.is_established

        assoc.dimse = DummyDIMSE()

        ds = Dataset()
        ds.PatientName = "Test^test"
        # Receives None, None from DummyDIMSE, aborts
        assoc.send_n_delete(
            Printer,
            "1.2.840.10008.5.1.1.40.1",
            meta_uid=BasicGrayscalePrintManagementMeta,
        )
        assert assoc.is_aborted

        scp.shutdown()

        assert assoc.dimse.req.RequestedSOPClassUID == Printer
        assert assoc.dimse.context_id == 1
        assert (
            assoc._accepted_cx[1].abstract_syntax == BasicGrayscalePrintManagementMeta
        )

    def test_meta_uid_good(self):
        """Test sending a request using a Meta SOP Class."""
        handler_data = []

        def handle(event):
            handler_data.append(event)
            return 0x0000

        self.ae = ae = AE()
        ae.acse_timeout = 5
        ae.dimse_timeout = 5
        ae.network_timeout = 5
        ae.add_supported_context(BasicGrayscalePrintManagementMeta)
        scp = ae.start_server(
            ("localhost", 11112), block=False, evt_handlers=[(evt.EVT_N_DELETE, handle)]
        )

        ae.add_requested_context(BasicGrayscalePrintManagementMeta)
        assoc = ae.associate("localhost", 11112)
        assert assoc.is_established

        status = assoc.send_n_delete(
            BasicFilmSession,
            "1.2.840.10008.5.1.1.40.1",
            meta_uid=BasicGrayscalePrintManagementMeta,
        )
        assert status.Status == 0x0000
        assoc.release()
        assert assoc.is_released

        scp.shutdown()

        req = handler_data[0].request
        cx = handler_data[0].context

        assert req.RequestedSOPClassUID == BasicFilmSession
        assert cx.abstract_syntax == BasicGrayscalePrintManagementMeta

    def test_meta_uid_bad(self):
        """Test sending a request using a Meta SOP Class."""

        def handle(event):
            ds = Dataset()
            ds.PatientName = "Test"
            ds.SOPClassUID = DisplaySystem
            ds.SOPInstanceUID = "1.2.3.4"
            return 0x0000, ds

        self.ae = ae = AE()
        ae.acse_timeout = 5
        ae.dimse_timeout = 5
        ae.network_timeout = 5
        ae.add_supported_context(BasicGrayscalePrintManagementMeta)
        scp = ae.start_server(
            ("localhost", 11112), block=False, evt_handlers=[(evt.EVT_N_DELETE, handle)]
        )

        ae.add_requested_context(BasicGrayscalePrintManagementMeta)
        assoc = ae.associate("localhost", 11112)
        assert assoc.is_established

        ds = Dataset()
        ds.PatientName = "Test^test"
        msg = (
            r"No presentation context for 'Basic Color Print Management "
            r"Meta SOP Class' has been "
            r"accepted by the peer "
            r"for the SCU role"
        )
        with pytest.raises(ValueError, match=msg):
            assoc.send_n_delete(
                BasicFilmSession,
                "1.2.840.10008.5.1.1.40.1",
                meta_uid=BasicColorPrintManagementMeta,
            )

        assoc.release()
        assert assoc.is_released

        scp.shutdown()
